#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02-12-2020
           """

import torch
from warg import kws_sink
from draugr.torch_utilities.system.device import global_torch_device
from draugr.torch_utilities.sessions.model_sessions import TorchEvalSession

from torch.utils.data import DataLoader
from tqdm import tqdm

__all__ = ["find_n_misclassified"]


def find_n_misclassified(
    model: torch.nn.Module,
    evaluation_loader: DataLoader,
    *,
    mapper: callable = kws_sink,
    n: int = 10,
    device: torch.device = global_torch_device(),
) -> None:
    j = 0
    num_samples = len(evaluation_loader)
    with TorchEvalSession(model):
        for i, (waveform, target) in tqdm(enumerate(evaluation_loader), total=n):
            output = mapper(model(waveform.to(device)).argmax(dim=-1).squeeze())
            truth = mapper(target)
            if output != truth:
                print(
                    f"Data point #{i}/{num_samples}. Expected: {truth}. Predicted: {output}."
                )
                j += 1
                if j >= n:
                    break
        else:
            print("All examples in this dataset were correctly classified!")
            print("In this case, let's just look at the last data point")
            print(f"Data point #{i}. Expected: {truth}. Predicted: {output}.")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25/03/2020
           """

import torch

__all__ = ["minmax_to_xywh_torch"]


def minmax_to_xywh_torch(boxes: torch.tensor) -> torch.tensor:
    """

  :param boxes:
  :type boxes:
  :return:
  :rtype:"""
    xmin, ymin, xmax, ymax = boxes.unbind(1)
    return torch.stack((xmin, ymin, xmax - xmin, ymax - ymin), dim=1)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

from abc import ABCMeta, abstractmethod
from collections import Counter, deque

__all__ = ["Writer", "global_writer", "set_global_writer"]

from itertools import cycle
from typing import Iterable

from warg import is_none_or_zero_or_negative_or_mod_zero
from warg import Number, drop_unused_kws


class Writer(metaclass=ABCMeta):
    """"""

    @drop_unused_kws
    def __init__(
        self, *, interval: int = 1, filters: Iterable = None, verbose: bool = False
    ):
        """

    :param interval:
    :param filters:
    :param verbose:
    """
        self._counter = Counter()
        self._blip_values = iter(cycle(range(2)))
        self._interval = interval
        self.filters = filters
        self._verbose = verbose

    def filter(self, tag: str) -> bool:
        """

        returns a boolean  value, true if to be included, False if to be excluded

        tag is in filter if not None
        and within interval for inclusion

    :param tag:
    :type tag:
    :return:
    :rtype:"""
        is_in_filters = self.filters is None or tag in self.filters
        at_interval = is_none_or_zero_or_negative_or_mod_zero(
            self._interval, self._counter[tag]
        )
        return is_in_filters and at_interval

    def __enter__(self):
        global GLOBAL_WRITER_STACK, GLOBAL_WRITER
        GLOBAL_WRITER_STACK.appendleft(self)
        GLOBAL_WRITER = self
        return self._open()

    def __exit__(self, exc_type, exc_val, exc_tb):
        global GLOBAL_WRITER, GLOBAL_WRITER_STACK

        if len(GLOBAL_WRITER_STACK) > 0:
            GLOBAL_WRITER_STACK.popleft()  # pop self

        if len(GLOBAL_WRITER_STACK) > 0:
            GLOBAL_WRITER = GLOBAL_WRITER_STACK.popleft()  # then previous
        else:
            GLOBAL_WRITER = None
        return self._close(exc_type, exc_val, exc_tb)

    def scalar(self, tag: str, value: Number, step_i: int = None) -> None:
        """

    :param tag:
    :type tag:
    :param value:
    :type value:
    :param step_i:
    :type step_i:"""
        if step_i:
            if self.filter(tag):
                self._scalar(tag, value, self._counter[tag])
            self._counter[tag] = step_i
        else:
            if self.filter(tag):
                self._scalar(tag, value, self._counter[tag])
            self._counter[tag] += 1

    def blip(self, tag: str, step_i: int = None) -> None:
        """

    :param tag:
    :type tag:
    :param step_i:
    :type step_i:"""
        if step_i:
            self.scalar(tag, next(self._blip_values), step_i)
            self.scalar(tag, next(self._blip_values), step_i)
        else:
            self.scalar(tag, next(self._blip_values))
            self.scalar(tag, next(self._blip_values), self._counter[tag])

    def close(self):
        """"""
        self._close()

    def open(self):
        """"""
        self._open()

    @abstractmethod
    def _scalar(self, tag: str, value: float, step: int):
        raise NotImplementedError

    @abstractmethod
    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        raise NotImplementedError

    @abstractmethod
    def _open(self):
        return self


GLOBAL_WRITER_STACK = deque()
GLOBAL_WRITER = None  #: Writer = None


def global_writer() -> Writer:
    """

  :return:
  :rtype:"""
    global GLOBAL_WRITER
    return GLOBAL_WRITER


def set_global_writer(writer: Writer) -> None:
    """

  :return:
  :rtype:"""
    global GLOBAL_WRITER
    # if GLOBAL_WRITER:
    # GLOBAL_WRITER_STACK TODO: push to stack if existing?

    GLOBAL_WRITER = writer



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import torch

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """

__all__ = [
    "uint_nhwc_to_nchw_float_batch",
    "float_nchw_to_nhwc_uint_batch",
    "hwc_to_chw_tensor",
    "chw_to_hwc_tensor",
    "float_chw_to_hwc_uint_tensor",
    "uint_hwc_to_chw_float_tensor",
]

from PIL.Image import Image


def hwc_to_chw_tensor(inp: torch.Tensor) -> torch.Tensor:
    """

  :param inp:
  :type inp:
  :return:
  :rtype:"""
    assert len(inp.shape) == 3
    return inp.permute(2, 0, 1)


def chw_to_hwc_tensor(inp: torch.Tensor) -> torch.Tensor:
    """

  :param inp:
  :type inp:
  :return:
  :rtype:"""
    assert len(inp.shape) == 3
    return inp.permute(1, 2, 0)


def uint_hwc_to_chw_float_tensor(
    inp: torch.Tensor, *, normalise: bool = True
) -> torch.Tensor:
    """

  :param inp:
  :type inp:
  :param normalise:
  :type normalise:
  :return:
  :rtype:"""
    if normalise:
        inp = (inp / 255.0).clamp(0, 1)
    return hwc_to_chw_tensor(inp)


def float_chw_to_hwc_uint_tensor(
    inp: torch.Tensor, *, unnormalise: bool = True
) -> torch.Tensor:
    """

  :param inp:
  :type inp:
  :param unnormalise:
  :type unnormalise:
  :return:
  :rtype:"""
    inp = chw_to_hwc_tensor(inp)
    if unnormalise:
        inp = (inp * 255.0).clamp(0, 255)
    return inp.to(dtype=torch.uint8)


def uint_nhwc_to_nchw_float_batch(
    inp: torch.Tensor, *, normalise: bool = True
) -> torch.Tensor:
    """

  :param inp:
  :type inp:
  :param normalise:
  :type normalise:
  :return:
  :rtype:"""
    assert len(inp.shape) == 4
    if normalise:
        inp = (inp / 255.0).clamp(0, 1)
    return inp.permute(0, 3, 1, 2)


def float_nchw_to_nhwc_uint_batch(
    inp: torch.Tensor, *, unnormalise: bool = True
) -> torch.Tensor:
    """

  :param inp:
  :type inp:
  :param unnormalise:
  :type unnormalise:
  :return:
  :rtype:"""
    assert len(inp.shape) == 4
    inp = inp.permute(0, 3, 1, 2)
    if unnormalise:
        inp = (inp * 255.0).clamp(0, 255)
    return inp.to(dtype=torch.uint8)


if __name__ == "__main__":
    hw = 2
    a = torch.ones(3, hw, hw)
    print(a)
    b = float_chw_to_hwc_uint_tensor(a)
    print(b)
    c = uint_hwc_to_chw_float_tensor(b)
    print(c)
    d = chw_to_hwc_tensor(c)
    assert (
        d.shape == c.T.shape
    )  # only work h and w is same size, mind that semantically not the same as transpose will be (whc)
    print(c.shape)
    print(d.shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import math

import torch
from draugr.torch_utilities.system.constants import torch_pi

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = [
    "shannon_entropy",
    "log_shannon_entropy",
    "normal_entropy",
    "differential_entropy_gaussian",
    "normal_log_density",
]


def shannon_entropy(prob: torch.Tensor) -> torch.Tensor:
    """

  :param prob:
  :type prob:
  :return:
  :rtype:"""
    return -torch.sum(prob * torch.log2(prob), -1)


def log_shannon_entropy(log_prob: torch.Tensor) -> torch.Tensor:
    """

  :param log_prob:
  :type log_prob:
  :return:
  :rtype:"""
    return -torch.sum(torch.pow(2, log_prob) * log_prob, -1)
    # return - torch.sum(torch.exp(log_prob) * log_prob, -1)


def normal_entropy(std: torch.Tensor) -> torch.Tensor:
    """

  :param std:
  :type std:
  :return:
  :rtype:"""
    var = std.pow(2)
    ent = 0.5 + 0.5 * torch.log(2 * var * math.pi)
    return ent.sum(dim=-1, keepdim=True)


def differential_entropy_gaussian(std: torch.Tensor) -> torch.Tensor:
    """

  :param std:
  :type std:
  :return:
  :rtype:"""
    return torch.log(std * torch.sqrt(2 * torch_pi())) + 0.5


def normal_log_density(x: torch.Tensor, mean, log_std, std) -> torch.Tensor:
    """

  :param x:
  :type x:
  :param mean:
  :type mean:
  :param log_std:
  :type log_std:
  :param std:
  :type std:
  :return:
  :rtype:"""
    var = std.pow(2)
    log_density = -(x - mean).pow(2) / (2 * var) - 0.5 * math.log(2 * math.pi) - log_std
    return log_density.sum(1, keepdim=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 10/07/2020
           """

from .module_writer_hook import *
from .module_writer_parameters import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import contextlib
from time import sleep
from typing import Iterable, MutableSequence

from draugr.python_utilities.styling import sprint

__author__ = "Christian Heider Nielsen"
__doc__ = ""

from warg import GDKC, drop_unused_kws, passes_kws_to
from pynput import keyboard

# import keyboard

__all__ = ["add_early_stopping_key_combination", "CaptureEarlyStop"]


@drop_unused_kws
def add_early_stopping_key_combination(
    callbacks: callable = None, has_x_server: bool = True, verbose: bool = False
) -> keyboard.Listener:
    """

:param callbacks:
:param has_x_server:
:param verbose:
:return:"""
    if not has_x_server:
        return

    COMBINATIONS = [
        {keyboard.Key.ctrl, keyboard.KeyCode(char="c")},
        {keyboard.Key.ctrl, keyboard.KeyCode(char="d")},
        {keyboard.Key.shift, keyboard.Key.alt, keyboard.KeyCode(char="s")},
        {keyboard.Key.shift, keyboard.Key.alt, keyboard.KeyCode(char="S")},
    ]

    CALLBACKS: MutableSequence = []
    # The currently active modifiers
    current = set()

    # keyboard.add_hotkey(key, callback)
    if callbacks is not None:
        if isinstance(callbacks, Iterable):
            CALLBACKS.extend(callbacks)
        else:
            CALLBACKS.append(callbacks)
    if verbose:
        sprint(
            f"\n\nPress any of:\n{COMBINATIONS}\n for early stopping\n",
            color="red",
            bold=True,
            highlight=True,
        )

    def on_press(key):
        if any([key in COMBO for COMBO in COMBINATIONS]):
            if verbose:
                print(f"Adding key {key}")
            current.add(key)
            if any(all(k in current for k in COMBO) for COMBO in COMBINATIONS):
                for clbck in CALLBACKS:
                    if verbose:
                        print(f"Calling {clbck}")
                    clbck()

    def on_release(key):
        if any([key in combo for combo in COMBINATIONS]):
            if key in current:
                if verbose:
                    print(f"Removing key {key}")
                current.remove(key)

    return keyboard.Listener(on_press=on_press, on_release=on_release)


class CaptureEarlyStop(contextlib.AbstractContextManager):
    """
Context for early stopping a loop"""

    @passes_kws_to(add_early_stopping_key_combination)
    def __init__(self, *args, **kwargs):
        self.listener = add_early_stopping_key_combination(*args, **kwargs)

    def __enter__(self):
        if self.listener:
            self.listener.start()

    def __exit__(self, exc_type, exc_value, traceback):
        if self.listener:
            self.listener.stop()
        return False


if __name__ == "__main__":

    def c():
        print("start")
        RUN = True

        def stop_loop():
            global RUN
            RUN = False

        with CaptureEarlyStop(stop_loop) as _:
            while RUN:
                sleep(0.1)
        print("done")

    def b():  # DOES NOT WORK!
        print("start2")
        with CaptureEarlyStop(GDKC(exit, code=0)) as _:
            while True:
                sleep(0.1)
        print("done2")

    c()
    # b()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20/03/2020
           """

import torch

from draugr.torch_utilities.sessions.device_sessions import (
    global_torch_device,
    TorchCpuSession,
    TorchCudaSession,
)
from draugr.torch_utilities.sessions.model_sessions import (
    TorchEvalSession,
    TorchTrainSession,
)

__all__ = ["TorchCacheSession"]

from warg import AlsoDecorator


class TorchCacheSession(AlsoDecorator):
    """
  # speed up evaluating after training finished
  # NOTE: HAS THE SIDE EFFECT OF CLEARING CACHE, NON RECOVERABLE
  """

    def __init__(self, using_cuda: bool = global_torch_device().type == "cuda"):
        self.using_cuda = using_cuda

    def __enter__(self):
        if self.using_cuda:
            torch.cuda.empty_cache()
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.using_cuda:
            torch.cuda.empty_cache()


if __name__ == "__main__":

    def a():
        print(torch.cuda.memory_cached(global_torch_device()))
        with TorchCacheSession():
            torch.tensor([0.0], device=global_torch_device())
            print(torch.cuda.memory_cached(global_torch_device()))
        print(torch.cuda.memory_cached(global_torch_device()))

    def b():
        model = torch.nn.Sequential(torch.nn.Linear(1, 1), torch.nn.Dropout(0.1))
        print(model.training)
        with TorchEvalSession(model):
            print(model.training)
        print(model.training)

    def c():
        model = torch.nn.Sequential(torch.nn.Linear(1, 1), torch.nn.Dropout(0.1))
        model.eval()
        print(model.training)
        with TorchTrainSession(model):
            print(model.training)
        print(model.training)

    def d():
        print(
            global_torch_device(
                override=global_torch_device(cuda_device_preference=False)
            )
        )
        print(global_torch_device())
        with TorchCudaSession():
            print(global_torch_device())
        print(global_torch_device())

    def e():
        print(
            global_torch_device(
                override=global_torch_device(cuda_device_preference=True)
            )
        )
        print(global_torch_device())
        with TorchCpuSession():
            print(global_torch_device())
        print(global_torch_device())

    # a()
    # b()
    # c()
    d()
    e()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import os
import shutil
from pathlib import Path
from random import randint
from time import time
from typing import Any, List, Mapping, Sequence, Union

import numpy
from warg import NOD, passes_kws_to

__author__ = "Christian Heider Nielsen"

import six

__all__ = [
    "COLORS",
    "DECORATIONS",
    "generate_style",
    "sprint",
    "PrintStyle",
    "scale",
    "get_terminal_size",
    "hyperlink_path",
    "hyperlink_url",
]

COLORS = NOD(
    red="31",
    green="32",
    yellow="33",
    # gray='30', #Black,
    blue="34",
    magenta="35",
    cyan="36",
    white="37",
    crimson="38",
)

DECORATIONS = NOD(
    end="0",
    bold="1",
    dim="2",
    italic="3",
    underline="4",
    underline_end="24",  # '4:0',
    double_underline="21",  # '4:2'
    # double_underline_end='24',  # '4:0'
    curly_underline="4:3",
    blink="5",
    reverse_colors="7",
    invisible="8",  # still copyable
    strikethrough="9",
    overline="53",
    hyperlink="8;;",
)


def hyperlink_url(
    link: str, link_id: str = None, style: bool = False, prefix: str = "https://"
) -> str:
    if not link.startswith(prefix):
        if "://" in link:
            link = f'{prefix}{link.split("://")[1]}'
        else:
            link = f"{prefix}{link}"

    if not style:
        return link
    if not link_id:
        link_id = f"{time()}-{randint(0, 999999)}" if link else ""
    return f"\x1b]8;id={link_id};{link}\x1b\\{link}\x1b]8;;\x1b\\"


def hyperlink_path(
    link: Path, link_id: str = None, style: bool = False, linux: bool = True
) -> str:
    if linux:
        link = f"file://{link.resolve()}"
    if not style:
        return link
    if not link_id:
        link_id = f"{time()}-{randint(0, 999999)}" if link else ""
    return f"\x1b]8;id={link_id};{link}\x1b\\{link}\x1b]8;;\x1b\\"


class PrintStyle(object):
    """"""

    def __init__(self, attributes_joined, end):
        self._attributes_joined = attributes_joined
        self._end = end

    def __call__(self, obj, *args, **kwargs) -> str:
        intermediate_repr = f"\x1b[{self._attributes_joined}m{obj}\x1b[{self._end}m"
        string = six.u(intermediate_repr)
        return string


def generate_style(
    obj: Any = None,
    *,
    color: Union[str, int] = "random",
    bold: bool = False,
    highlight: bool = False,
    underline: bool = False,
    italic: bool = False,
) -> Union[str, PrintStyle]:
    """

  :param obj:
  :type obj:
  :param color:
  :type color:
  :param bold:
  :type bold:
  :param highlight:
  :type highlight:
  :param underline:
  :type underline:
  :param italic:
  :type italic:
  :return:
  :rtype:"""
    attributes = []

    if color == "random":
        num = int(list(COLORS.values())[randint(0, len(COLORS) - 1)])
    elif color in COLORS:
        num = int(COLORS[color])
    elif color in COLORS.values():
        num = int(color)
    elif isinstance(color, int) and color < 39:
        num = color
    else:
        num = int(COLORS["white"])

    if highlight:
        num += 10

    attributes.append(six.u(f"{num}"))

    if bold:
        attributes.append(six.u(f'{DECORATIONS["bold"]}'))

    if underline:
        attributes.append(six.u(f'{DECORATIONS["underline"]}'))

    if italic:
        attributes.append(six.u(f'{DECORATIONS["italic"]}'))

    end = DECORATIONS["end"]

    attributes_joined = six.u(";").join(attributes)

    print_style = PrintStyle(attributes_joined, end)

    if obj:
        return print_style(obj)
    else:
        return print_style


@passes_kws_to(generate_style)
def sprint(obj: Any, print_kws: Mapping = {}, **kwargs) -> None:
    """
  Stylised print. Defaults to stdout
  Valid colors: gray, red, green, yellow, blue, magenta, cyan, white, crimson"""
    print(generate_style(obj, **kwargs), **print_kws)


def scale(x: Sequence, length: float) -> List[int]:
    """
  Scale points in 'x', such that distance between
  max(x) and min(x) equals to 'length'. min(x)
  will be moved to 0."""
    if type(x) is list:
        s = float(length) / (max(x) - min(x)) if x and max(x) - min(x) != 0 else length
    # elif type(x) is range:
    #  s = length
    else:
        s = (
            float(length) / (numpy.max(x) - numpy.min(x))
            if len(x) and numpy.max(x) - numpy.min(x) != 0
            else length
        )

    return [int((i - min(x)) * s) for i in x]


def get_terminal_size() -> NOD:
    """

  :return:
  :rtype:"""
    try:
        size = shutil.get_terminal_size()
        columns, rows = size.columns, size.lines
    except:
        rows, columns = (os.getenv("LINES", 25), os.getenv("COLUMNS", 80))

    rows, columns = int(rows), int(columns)

    return NOD(rows=rows, columns=columns)


if __name__ == "__main__":
    sprint(get_terminal_size())
    sprint(hyperlink_path(Path.home()))
    print(hyperlink_path(Path.home()))
    sprint(hyperlink_path(Path.home() / ".xprofile"))
    sprint(hyperlink_url("http://dr.dk"))
    sprint(hyperlink_url("dr.dk"))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import logging
import pathlib
import sys

from apppath import ensure_existence
from draugr import PROJECT_APP_PATH
from draugr.writers.writer import Writer

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""
__all__ = ["LogWriter"]


class LogWriter(Writer):
    """"""

    def _scalar(self, tag: str, value: float, step: int) -> None:
        self.logger.info(f"{step} [{tag}] {value}")

    @staticmethod
    def get_logger(
        path: pathlib.Path = pathlib.Path.cwd() / "0.log",
        write_to_std_out: bool = False,
    ) -> logging.Logger:
        """

    :param path:
    :type path:
    :param write_to_std_out:
    :type write_to_std_out:
    :return:
    :rtype:"""
        ensure_existence(path, declare_file=True, overwrite_on_wrong_type=True)

        handlers = [logging.FileHandler(filename=str(path))]

        if write_to_std_out:
            handlers.append(logging.StreamHandler(sys.stdout))

        logging.basicConfig(level=logging.INFO, format="%(message)s", handlers=handlers)

        return logging.getLogger()

    def __init__(self, path, **kwargs):
        super().__init__(**kwargs)
        self.log_path = path
        self.logger: logging.Logger = None

    def _open(self):
        self.logger = self.get_logger(self.log_path)
        return self

    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        del self.logger

    def __getattr__(self, item):
        return getattr(self.logger, item)

    def __call__(self, msg):
        self.logger.info(msg)


if __name__ == "__main__":

    with LogWriter(PROJECT_APP_PATH.user_log / "test") as w:
        w.scalar("What", 4)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 13-01-2021
           """

from math import ceil, log, log10, log2, floor

__all__ = [
    "next_pow",
    "next_power",
    "next_pow_2",
    "next_pow_10",
    "next_power_2",
    "next_power_10",
    "prev_pow",
    "prev_pow_10",
    "prev_pow_2",
]

from warg import Number


def next_pow(x: Number, n: int = None) -> int:
    """
  If the base(n) not specified, returns the natural logarithm (base e) of x.

  @param x:
  @param n:
  @return:
  """
    return int(pow(n, ceil(log(x, n))))


def prev_pow(x: Number, n: int = None) -> int:
    """
  If the base(n) not specified, returns the natural logarithm (base e) of x.

  @param x:
  @param n:
  @return:
  """
    return int(pow(n, floor(log(x, n))))


def next_pow_2(x: Number) -> int:
    return int(pow(2, ceil(log2(x))))


def prev_pow_2(x: Number) -> int:
    return int(pow(2, floor(log2(x))))


def next_pow_10(x: Number) -> int:
    return int(pow(10, ceil(log10(x))))


def prev_pow_10(x: Number) -> int:
    return int(pow(10, floor(log10(x))))


next_power = next_pow
next_power_2 = next_pow_2
next_power_10 = next_pow_10

if __name__ == "__main__":
    print(next_pow(17, 5))
    print(prev_pow(17, 5))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/01/2020
           """

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .bounding_boxes import *
from .color_space import *
from .cv2_transforms import *
from .frames import *
from .opencv_draw import *
from .opencv_drawing_utilities import *
from .resize import *
from .async_video_stream import *



import torch
from torch.utils.data import TensorDataset

from draugr.torch_utilities import cross_validation_generator, to_tensor


def asdasidoj():
    X = to_tensor([torch.diag(torch.arange(i, i + 2)) for i in range(200)])
    x_train = TensorDataset(X[:100])
    x_val = TensorDataset(X[100:])

    for train, val in cross_validation_generator(x_train, x_val):
        print(len(train), len(val))
        print(train[0], val[0])


asdasidoj()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

from functools import reduce

import numpy
from matplotlib import patches, pyplot

__all__ = [
    "denormalise_minusoneone",
    "matplotlib_bounding_box",
    "plot_img_array",
    "plot_side_by_side",
    "remove_decoration",
]

from matplotlib.axes import Axes


def remove_decoration(ax_: Axes) -> None:
    transparent = (1.0, 1.0, 1.0, 0.0)

    ax_.w_xaxis.set_pane_color(transparent)
    ax_.w_yaxis.set_pane_color(transparent)
    ax_.w_zaxis.set_pane_color(transparent)

    ax_.w_xaxis.line.set_color(transparent)
    ax_.w_yaxis.line.set_color(transparent)
    ax_.w_zaxis.line.set_color(transparent)

    ax_.set_xticks([])
    ax_.set_yticks([])
    ax_.set_zticks([])


def denormalise_minusoneone(T, coords):
    return 0.5 * ((coords + 1.0) * T)


def matplotlib_bounding_box(x, y, size, color: str = "w"):
    x = int(x - (size / 2))
    y = int(y - (size / 2))
    rect = patches.Rectangle(
        (x, y), size, size, linewidth=1, edgecolor=color, fill=False
    )
    return rect


def plot_img_array(img_array: numpy.ndarray, n_col: int = 3) -> None:
    """

  :param img_array:
  :type img_array:
  :param n_col:
  :type n_col:
  :return:
  :rtype:"""
    n_row = len(img_array) // n_col

    f, plots = pyplot.subplots(
        n_row, n_col, sharex="all", sharey="all", figsize=(n_col * 4, n_row * 4)
    )

    for i in range(len(img_array)):
        plots[i // n_col, i % n_col].imshow(img_array[i])


def plot_side_by_side(img_arrays) -> None:
    flatten_list = reduce(lambda x, y: x + y, zip(*img_arrays))

    plot_img_array(numpy.array(flatten_list), n_col=len(img_arrays))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 08-12-2020
           """

import time

import numpy
import torch
from draugr import WorkerSession, batched_recycle
from draugr.torch_utilities import to_tensor_generator


def test_d3():
    channels_in = 3
    channels_out = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, channels_in, 512, 512)

    model = torch.nn.Sequential(
        torch.nn.Conv2d(channels_in, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
    ).to(device)

    generator = to_tensor_generator(
        batched_recycle(numpy.random.sample(data_shape), batch_size), device=device
    )

    with WorkerSession(0.3) as num_workers:
        dataloader = torch.utils.data.DataLoader(
            numpy.random.sample(data_shape),
            batch_size=batch_size,
            shuffle=True,
            num_workers=num_workers,
            pin_memory=True,
        )

        for _ in range(samples):
            s1 = time.time()
            for _, a in zip(range(batches), dataloader):
                model(a.to(device, dtype=torch.float))

            s2 = time.time()
            for _, a in zip(range(batches), generator):
                model(a)

            s3 = time.time()

            print(f"dataloader: {s2 - s1}")
            print(f"generator: {s3 - s2}")


if __name__ == "__main__":
    test_d3()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Sequence

from draugr.drawers.mpl_drawers.mpldrawer import MplDrawer
from warg import passes_kws_to

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 22/11/2019

           """

from matplotlib import pyplot

import numpy

__all__ = ["DistributionPlot"]


class DistributionPlot(MplDrawer):
    """"""

    @passes_kws_to(MplDrawer.__init__)
    @passes_kws_to(pyplot.hist)
    def __init__(
        self,
        window_length: int = None,
        title: str = "",
        data_label: str = "Magnitude",
        reverse: bool = False,
        overwrite: bool = False,
        render: bool = True,
        **kwargs
    ):
        super().__init__(render=render, **kwargs)

        if not render:
            return

        self.fig = pyplot.figure(figsize=(4, 4))

        self.overwrite = overwrite
        self.reverse = reverse
        self.window_length = window_length

        self.array = []
        self.hist_kws = kwargs
        pyplot.hist(self.array, **self.hist_kws)

        pyplot.ylabel(data_label)

        pyplot.title(title)
        pyplot.tight_layout()

    def _draw(self, data: Sequence):
        """

    :param data:
    :return:"""
        if not isinstance(data, Sequence):
            data = [data]

        self.array.extend(data)
        pyplot.cla()
        pyplot.hist(self.array, **self.hist_kws)


if __name__ == "__main__":
    delta = 1.0 / 60.0

    s = DistributionPlot()
    for _ in range(100):
        s.draw(numpy.random.sample(), delta)



from pathlib import Path

from apppath import ensure_existence
from matplotlib import pyplot

from draugr import PROJECT_APP_PATH
from draugr.tensorboard_utilities import TensorboardEventExporter
from draugr.writers.standard_tags import TrainingScalars

if __name__ == "__main__":
    save = False
    event_files = list(PROJECT_APP_PATH.user_log.rglob("events.out.tfevents.*"))
    if len(event_files) > 0:
        for _path_to_events_file in event_files:
            print(f"Event file: {_path_to_events_file}")
            _out_dir = Path.cwd() / "exclude" / "results"
            ensure_existence(_out_dir)
            tee = TensorboardEventExporter(
                _path_to_events_file.parent, save_to_disk=save
            )
            print(f"Available tags: {tee.tags_available}")
            tee.export_line_plot(TrainingScalars.training_loss.value, out_dir=_out_dir)
            if not save:
                pyplot.show()
    else:
        print("No events found")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import time

import numpy
import torch
from torch.utils.data import Dataset

from draugr import batched_recycle
from draugr.torch_utilities import to_tensor_generator

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/10/2019
           """


def test_d1():
    channels_in = 3
    channels_out = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, channels_in, 512, 512)

    model = torch.nn.Sequential(
        torch.nn.Conv2d(channels_in, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
    ).to(device)

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(
            range(batches),
            to_tensor_generator(
                batched_recycle(numpy.random.sample(data_shape), batch_size),
                device=device,
                preload_next=False,
            ),
        ):
            model(a)

        s2 = time.time()
        for _, a in zip(
            range(batches),
            torch.utils.data.DataLoader(
                numpy.random.sample(data_shape),
                batch_size=batch_size,
                shuffle=True,
                num_workers=1,
                pin_memory=False,
            ),
        ):
            model(a.to(device, dtype=torch.float))

        s3 = time.time()

        print(f"generator: {s2 - s1}")
        print(f"dataloader: {s3 - s2}")


def test_d2():
    channels_in = 3
    channels_out = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, channels_in, 512, 512)

    model = torch.nn.Sequential(
        torch.nn.Conv2d(channels_in, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
    ).to(device)

    generator = to_tensor_generator(
        batched_recycle(numpy.random.sample(data_shape), batch_size),
        device=device,
        preload_next=False,
    )

    dataloader = torch.utils.data.DataLoader(
        numpy.random.sample(data_shape),
        batch_size=batch_size,
        shuffle=True,
        num_workers=1,
        pin_memory=False,
    )

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(range(batches), generator):
            model(a)

        s2 = time.time()
        for _, a in zip(range(batches), dataloader):
            model(a.to(device, dtype=torch.float))
        s3 = time.time()

        print(f"generator: {s2 - s1}")
        print(f"dataloader: {s3 - s2}")


def test_d3():
    channels_in = 3
    channels_out = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, channels_in, 512, 512)

    model = torch.nn.Sequential(
        torch.nn.Conv2d(channels_in, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
    ).to(device)

    generator = to_tensor_generator(
        batched_recycle(numpy.random.sample(data_shape), batch_size), device=device
    )

    dataloader = torch.utils.data.DataLoader(
        numpy.random.sample(data_shape),
        batch_size=batch_size,
        shuffle=True,
        num_workers=4,
        pin_memory=True,
    )

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(range(batches), dataloader):
            model(a.to(device, dtype=torch.float))

        s2 = time.time()
        for _, a in zip(range(batches), generator):
            model(a)

        s3 = time.time()

        print(f"dataloader: {s2 - s1}")
        print(f"generator: {s3 - s2}")


def test_d4():
    from torchvision.transforms import transforms
    import numpy
    from draugr.generators.recycling import batched_recycle
    from draugr import inner_map

    a_transform = transforms.Compose(
        [
            transforms.ToPILImage("RGB"),
            transforms.Resize(224),
            transforms.CenterCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    channels_in = 3
    channels_out = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, 256, 256, channels_in)
    batch_shape = torch.Size([batch_size, channels_in, 224, 224])

    model = torch.nn.Sequential(
        torch.nn.Conv2d(channels_in, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
    ).to(device)

    class RandomDataset(Dataset):
        def __init__(self):
            self.d = numpy.random.sample(data_shape)

        def __len__(self):
            return len(self.d)

        def __getitem__(self, item):
            return a_transform(self.d[item])

    dataloader = torch.utils.data.DataLoader(
        RandomDataset(),
        batch_size=batch_size,
        shuffle=True,
        num_workers=1,
        pin_memory=False,
    )

    generator = to_tensor_generator(
        inner_map(
            a_transform, batched_recycle(numpy.random.sample(data_shape), batch_size)
        ),
        device=device,
    )

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(range(batches), dataloader):
            assert batch_shape == a.shape, a.shape
            model(a.to(device, dtype=torch.float))

        s2 = time.time()
        for _, a in zip(range(batches), generator):
            assert batch_shape == a.shape, a.shape
            model(a)

        s3 = time.time()

        print(f"dataloader: {s2 - s1}")
        print(f"generator: {s3 - s2}")


def test_d5():
    from torchvision.transforms import transforms
    import numpy
    from draugr.generators.recycling import batched_recycle
    from draugr import inner_map

    a_transform = transforms.Compose(
        [
            transforms.ToPILImage("RGB"),
            transforms.Resize(224),
            transforms.CenterCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    channels_in = 3
    channels_out = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, 256, 256, channels_in)
    batch_shape = torch.Size([batch_size, channels_in, 224, 224])

    class RandomDataset(Dataset):
        def __init__(self):
            self.d = numpy.random.sample(data_shape)

        def __len__(self):
            return len(self.d)

        def __getitem__(self, item):
            return a_transform(self.d[item])

    dataloader = torch.utils.data.DataLoader(
        RandomDataset(),
        batch_size=batch_size,
        shuffle=True,
        num_workers=1,
        pin_memory=False,
    )

    generator = to_tensor_generator(
        inner_map(
            a_transform, batched_recycle(numpy.random.sample(data_shape), batch_size)
        ),
        device=device,
    )

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(range(batches), generator):
            assert batch_shape == a.shape, a.shape

        s2 = time.time()

        for _, a in zip(range(batches), dataloader):
            assert batch_shape == a.shape, a.shape
        s3 = time.time()

        print(f"generator: {s2 - s1}")
        print(f"dataloader: {s3 - s2}")


def test_d6():
    from torchvision.transforms import transforms
    import numpy
    from draugr.generators.recycling import batched_recycle
    from draugr import inner_map

    a_transform = transforms.Compose(
        [
            transforms.ToPILImage("RGB"),
            transforms.Resize(224),
            transforms.CenterCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
        ]
    )

    channels_in = 3
    channels_out = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, 256, 256, channels_in)
    batch_shape = torch.Size([batch_size, channels_in, 224, 224])

    class RandomDataset(Dataset):
        def __init__(self):
            self.d = numpy.random.sample(data_shape)

        def __len__(self):
            return len(self.d)

        def __getitem__(self, item):
            return a_transform(self.d[item])

    dataloader = torch.utils.data.DataLoader(
        RandomDataset(),
        batch_size=batch_size,
        shuffle=True,
        num_workers=1,
        pin_memory=True,
    )

    generator = to_tensor_generator(
        inner_map(
            a_transform, batched_recycle(numpy.random.sample(data_shape), batch_size)
        ),
        device=device,
        preload_next=True,
    )

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(range(batches), generator):
            assert batch_shape == a.shape, a.shape

        s2 = time.time()

        for _, a in zip(range(batches), dataloader):
            assert batch_shape == a.shape, a.shape
        s3 = time.time()

        print(f"generator: {s2 - s1}")
        print(f"dataloader: {s3 - s2}")


def test_d7():
    import numpy
    from draugr.generators.recycling import batched_recycle

    channels_in = 3

    samples = 10
    device = "cuda"
    batches = 3
    batch_size = 32
    data_shape = (batches * batch_size, 256, 256, channels_in)
    batch_shape = torch.Size([batch_size, 256, 256, channels_in])
    dtype = torch.float

    class RandomDataset(Dataset):
        def __init__(self):
            self.d = numpy.random.sample(data_shape)

        def __len__(self):
            return len(self.d)

        def __getitem__(self, item):
            return self.d[item]

    dataloader = torch.utils.data.DataLoader(
        RandomDataset(),
        batch_size=batch_size,
        shuffle=True,
        num_workers=1,
        pin_memory=True,
    )

    generator = to_tensor_generator(
        batched_recycle(numpy.random.sample(data_shape), batch_size),
        device=device,
        preload_next=True,
        dtype=dtype,
    )

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(range(batches), generator):
            assert batch_shape == a.shape, a.shape

        s2 = time.time()

        for _, a in zip(range(batches), dataloader):
            a = a.to(device, dtype=dtype)
            assert batch_shape == a.shape, a.shape
        s3 = time.time()

        print(f"generator: {s2 - s1}")
        print(f"dataloader: {s3 - s2}")


if __name__ == "__main__":
    test_d7()



from enum import Enum

__all__ = [
    "TrainingScalars",
    "TrainingCurves",
    "TrainingTables",
    "TestingScalars",
    "TestingCurves",
    "TestingTables",
]


class TrainingScalars(Enum):
    training_loss = "training_loss"
    validation_loss = "validation_loss"
    validation_accuracy = "validation_accuracy"
    new_best_model = "new_best_model"


class TrainingCurves(Enum):
    validation_precision_recall = "validation_precision_recall"
    validation_receiver_operator_characteristic = (
        "validation_receiver_operator_characteristic"
    )


class TrainingTables(Enum):
    validation_confusion_matrix = "validation_confusion_matrix"
    validation_support = "validation_support"


class TestingScalars(Enum):
    test_accuracy = "test_accuracy"
    test_precision = "test_precision"
    test_recall = "test_recall"
    test_receiver_operator_characteristic_auc = (
        "test_receiver_operator_characteristic_auc"
    )


class TestingCurves(Enum):
    test_precision_recall = "test_precision_recall"
    test_receiver_operator_characteristic = "test_receiver_operator_characteristic"


class TestingTables(Enum):
    test_confusion_matrix = "test_confusion_matrix"
    test_support = "test_support"


if __name__ == "__main__":
    print(TrainingScalars.training_loss.value)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Union

import numpy
import torch
from PIL import Image

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["AudioWriterMixin"]


class AudioWriterMixin(ABC):
    """
  Writer subclass that provides an interface for 'writing' images"""

    @abstractmethod
    def audio(
        self,
        tag: str,
        data: Union[numpy.ndarray, torch.Tensor, Image.Image],
        step,
        **kwargs
    ) -> None:
        """

    :param tag:
    :type tag:
    :param data:
    :type data:
    :param step:
    :type step:
    :param kwargs:
    :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import datetime
import os
from pathlib import Path
from logging import warning

import pkg_resources
from apppath import AppPath

__project__ = "Draugr"
__author__ = "Christian Heider Nielsen"
__version__ = "0.8.4"
__doc__ = """
Created on 27/04/2019

@author: cnheider

"""

with open(Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

# __all__ = ["PROJECT_APP_PATH", "PROJECT_NAME", "PROJECT_VERSION", "get_version"]


from typing import Any

# from .drawers import *
# from .writers import *
# from .opencv_utilities import *
# from .torch_utilities import *
# from .stopping import *
# from .numpy_utilities import *
from .visualisation import *
from .metrics import *
from .generators import *
from .python_utilities import *


def dist_is_editable(dist: Any) -> bool:
    """
  Return True if given Distribution is an editable install."""
    import sys

    for path_item in sys.path:
        egg_link = Path(path_item) / f"{dist.project_name}.egg-link"
        if egg_link.is_file():
            return True
    return False


PROJECT_NAME = __project__.lower().strip().replace(" ", "_")
PROJECT_VERSION = __version__
PROJECT_YEAR = 2018
PROJECT_AUTHOR = __author__.lower().strip().replace(" ", "_")
PROJECT_APP_PATH = AppPath(app_name=PROJECT_NAME, app_author=PROJECT_AUTHOR)
PACKAGE_DATA_PATH = Path(pkg_resources.resource_filename(PROJECT_NAME, "data"))
INCLUDE_PROJECT_READMES = False

distributions = {v.key: v for v in pkg_resources.working_set}
if PROJECT_NAME in distributions:
    distribution = distributions[PROJECT_NAME]
    DEVELOP = dist_is_editable(distribution)
else:
    DEVELOP = True


def get_version(append_time: Any = DEVELOP) -> str:
    version = __version__
    if not version:
        version = os.getenv("VERSION", "0.0.0")

    if append_time:
        now = datetime.datetime.utcnow()
        date_version = now.strftime("%Y%m%d%H%M%S")
        # date_version = time.time()

        if version:
            # Most git tags are prefixed with 'v' (example: v1.2.3) this is
            # never desirable for artifact repositories, so we strip the
            # leading 'v' if it's present.
            version = (
                version[1:]
                if isinstance(version, str) and version.startswith("v")
                else version
            )
        else:
            # Default version is an ISO8601 compliant datetime. PyPI doesn't allow
            # the colon ':' character in its versions, and time is required to allow
            # for multiple publications to master in one day. This datetime string
            # uses the 'basic' ISO8601 format for both its date and time components
            # to avoid issues with the colon character (ISO requires that date and
            # time components of a date-time string must be uniformly basic or
            # extended, which is why the date component does not have dashes.
            #
            # Publications using datetime versions should only be made from master
            # to represent the HEAD moving forward.
            warning(
                f"Environment variable VERSION is not set, only using datetime: {date_version}"
            )

            # warn(f'Environment variable VERSION is not set, only using timestamp: {version}')

        version = f"{version}.{date_version}"

    return version


if __version__ is None:
    __version__ = get_version(append_time=True)

__version_info__ = tuple(int(segment) for segment in __version__.split("."))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import numpy
import torch

from draugr.torch_utilities.tensors.to_tensor import to_tensor

__author__ = "Christian Heider Nielsen"
__doc__ = ""


def test_to_tensor_none():
    try:
        tensor = to_tensor(None)
    except:
        return
    assert False


def test_to_tensor_empty_list():
    try:
        tensor = to_tensor([])
    except:
        return
    assert False


def test_to_tensor_empty_tuple():
    try:
        tensor = to_tensor(())
    except:
        return
    assert False


def test_to_tensor_list():
    ref = [0]
    tensor = to_tensor(ref, device="cpu")
    assert tensor.equal(torch.FloatTensor([0]))


def test_to_tensor_multi_list():
    ref = [[0], [1]]
    tensor = to_tensor(ref, device="cpu")
    assert tensor.equal(torch.FloatTensor([[0], [1]]))


def test_to_tensor_tuple():
    ref = (0,)
    tensor = to_tensor(ref, device="cpu")
    assert tensor.equal(torch.FloatTensor([0]))


def test_to_tensor_multi_tuple():
    ref = ([0], [1])
    tensor = to_tensor(ref, device="cpu")
    assert tensor.equal(torch.FloatTensor([[0], [1]]))


def test_to_tensor_from_numpy_tensor():
    ref = torch.from_numpy(numpy.random.sample((1, 2)))
    tensor = to_tensor(ref, dtype=torch.double, device="cpu")
    assert tensor.equal(ref)


def test_to_tensor_float_tensor():
    ref = torch.FloatTensor([0])
    tensor = to_tensor(ref, device="cpu")
    assert tensor.equal(ref)


def test_generator_to_float_tensor():
    s = range(9)
    ref = torch.FloatTensor([*s])
    tensor = to_tensor(s, device="cpu")
    assert tensor.equal(ref)


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

__all__ = ["replicate"]

from typing import Sequence, Union

from warg import Number


def replicate(x: Union[Sequence, Number], times: int = 2) -> Sequence:
    """
  if not tuple

  :param times:
  :type times:
  :param x:
  :type x:
  :return:
  :rtype:"""
    if isinstance(x, Sequence):
        if len(x) == times:
            return x
    return (x,) * times


if __name__ == "__main__":

    def asdaa():

        print(replicate(2))
        print(replicate(2, 4))

        print(replicate((2, 3)))
        print(replicate((2, 3), times=4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 06/04/2020
           """

from pathlib import Path

import numpy
from draugr.drawers import FastFourierTransformPlot, FastFourierTransformSpectrogramPlot
from matplotlib import pyplot
from scipy.io import wavfile

if __name__ == "__main__":

    def main():
        sampling_rate, audio = wavfile.read(
            str(Path.home() / "Data" / "Audio" / "Nightingale-sound.wav")
        )
        audio = numpy.mean(audio, axis=1)
        num_samples = audio.shape[0]
        length_sec = num_samples / sampling_rate

        print(f"Audio length: {length_sec:.2f} seconds")

        def a():
            n_fft = 32
            s = FastFourierTransformPlot(n_fft=n_fft, sampling_rate=sampling_rate)
            for sample in audio:
                s.draw(sample)

        def b():
            n_fft = 32
            delta = 1 / sampling_rate
            s = FastFourierTransformSpectrogramPlot(
                n_fft=n_fft,
                sampling_rate=sampling_rate,
                buffer_size_sec=delta * n_fft * 4,
            )
            for sample in audio:
                s.draw(sample)

        def c():
            f, ax = pyplot.subplots()
            ax.plot(numpy.arange(num_samples) / sampling_rate, audio)
            ax.set_xlabel("Time [s]")
            ax.set_ylabel("Amplitude [unknown]")
            pyplot.show()

        def d():
            from skimage import util

            M = 1024

            slices = util.view_as_windows(audio, window_shape=(M,), step=100)
            print(f"Audio shape: {audio.shape}, Sliced audio shape: {slices.shape}")

            win = numpy.hanning(M + 1)[:-1]
            slices = slices * win

            slices = slices.T
            print("Shape of `slices`:", slices.shape)

            spectrum = numpy.fft.fft(slices, axis=0)[: M // 2 + 1 : -1]
            spectrum = numpy.abs(spectrum)

            f, ax = pyplot.subplots(figsize=(4.8, 2.4))

            S = numpy.abs(spectrum)
            S = 20 * numpy.log10(S / numpy.max(S))

            ax.imshow(
                S,
                origin="lower",
                cmap="viridis",
                extent=(0, length_sec, 0, sampling_rate / 2 / 1000),
            )
            ax.axis("tight")
            ax.set_ylabel("Frequency [kHz]")
            ax.set_xlabel("Time [s]")
            pyplot.show()

        d()

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import random
from typing import Iterable, Sequence, Any

import numpy

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/10/2019
           """

__all__ = ["recycle", "batched_recycle"]


def recycle(iterable: Iterable) -> Any:
    """
  loops an iterable like itertools.cycle, but in a random order (Permutation) everytime the iterable is
  exhausted

  :param iterable:
  :return:"""
    if not isinstance(iterable, Sequence):
        iterable = list(iterable)
    while True:
        for element in random.sample(iterable, len(iterable)):
            yield element


def batched_recycle(sized: Sequence, batch_size: int = 32) -> Any:
    """Batches and re-cycles an array with a different permutation"""
    if isinstance(sized, numpy.ndarray):
        while True:
            yield sized[numpy.random.choice(sized.shape[0], batch_size, replace=False)]
    else:
        if not isinstance(sized, Sequence):
            sized = [*sized]
        while True:
            yield random.sample(sized, batch_size)


if __name__ == "__main__":
    a = [0, 1.0, 3.0, 2.2]
    for b in zip(range(3), batched_recycle(a, 2)):
        print(b)

    for b in zip(range(9), recycle(a)):
        print(b)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from shutil import rmtree

__author__ = "Christian Heider Nielsen"
__doc__ = ""


def main(keep_alive: bool = True) -> str:
    """

  :param keep_alive:
  :type keep_alive:
  :return:
  :rtype:"""
    from draugr.torch_utilities import launch_tensorboard
    from contextlib import suppress
    from time import sleep

    from apppath import AppPath

    import argparse

    parser = argparse.ArgumentParser(description="Option for launching tensorboard")
    parser.add_argument("NAME", type=str, help="App name to open AppPath for")
    parser.add_argument(
        "--author", type=str, help="App author to open AppPath for", default=None
    )
    parser.add_argument(
        "--version", type=str, help="App version to open AppPath for", default=None
    )
    parser.add_argument(
        "--clean",
        action="store_true",
        default=False,
        help="Adding --clean argument will wipe tensorboard logs",
    )
    parser.add_argument(
        "--port", default=6006, help="Which port should tensorboard be served on"
    )
    args = parser.parse_args()

    PROJECT_APP_PATH = AppPath(args.NAME, args.author, args.version)

    log_dir = PROJECT_APP_PATH.user_log

    if args.clean:
        print(f"Wiping {log_dir}")
        if log_dir.exists():
            rmtree(str(log_dir))
        else:
            log_dir.mkdir()

    address = launch_tensorboard(log_dir, args.port)

    if keep_alive:
        print(f"tensorboard address: {address} for log_dir {log_dir}")
        with suppress(KeyboardInterrupt):
            while True:
                sleep(10)

    return address


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/03/2020
           """

from abc import abstractmethod
from typing import Sequence

from draugr.torch_utilities.datasets.supervised.supervised_dataset import (
    SupervisedDataset,
)

__all__ = ["CategoricalDataset"]

from warg import OrderedSet


class CategoricalDataset(SupervisedDataset):
    """
  Categorical Dataset for discrete learning problems.
"""

    @property
    @abstractmethod
    def categories(self) -> OrderedSet[str]:
        raise NotImplementedError

    def idx_to_str(self, idx: Sequence[int]) -> Sequence[str]:
        return self.categories[idx]

    def str_to_idx(self, s: str) -> int:
        return self.categories.index(s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 27-01-2021
           """

__all__ = ["max_decimation_subsample", "fir_subsample", "fft_subsample"]

from typing import Tuple

import numpy
from scipy.signal import decimate, resample


def max_decimation_subsample(
    signal: numpy.ndarray, decimation_factor: int = 10
) -> numpy.ndarray:
    return numpy.max(signal.reshape(-1, decimation_factor), axis=1)


def fir_subsample(
    signal: numpy.ndarray, max_resolution: int, sampling_rate: int
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    if signal.shape[-1] > max_resolution:
        sub_signal = decimate(signal, (signal.shape[-1] // max_resolution) + 1, axis=-1)
    else:
        sub_signal = signal
    sub_time = numpy.linspace(
        0,
        signal.shape[-1] // sampling_rate,  # Get time from indices
        num=sub_signal.shape[-1],
    )
    return sub_time, sub_signal


def fft_subsample(
    signal: numpy.ndarray, max_resolution: int, sampling_rate: int
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    if signal.shape[-1] > max_resolution:
        sub_signal = resample(signal, max_resolution, axis=-1)
    else:
        sub_signal = signal
    sub_time = numpy.linspace(
        0,
        signal.shape[-1] // sampling_rate,  # Get time from indices
        num=sub_signal.shape[-1],
    )
    return sub_time, sub_signal


if __name__ == "__main__":

    def aisjd():
        a = numpy.arange(0, 100)
        print(max_decimation_subsample(a))

    aisjd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18-01-2021
           """

from enum import Enum
from typing import Iterable

__all__ = ["FilterModeEnum", "symbol_filter"]


class FilterModeEnum(Enum):
    exclude_postfix = "exclude_postfix"
    exclude_prefix = "exclude_prefix"
    exclude_fully = "exclude_fully"
    # TODO: Include variants


def symbol_filter(
    string_stream: Iterable,
    symbol: str = "#",
    *,
    exclusion_mode: FilterModeEnum = FilterModeEnum.exclude_postfix,
):
    if exclusion_mode == FilterModeEnum.exclude_fully:
        yield from filter(lambda s: symbol not in s, string_stream)
    elif (
        exclusion_mode == FilterModeEnum.exclude_postfix
        or exclusion_mode == FilterModeEnum.exclude_prefix
    ):
        selector = 0
        if exclusion_mode == FilterModeEnum.exclude_prefix:
            selector = -1
        for s in string_stream:
            raw = s.split(symbol)[selector].strip()
            if raw:
                yield raw
    else:
        raise NotImplemented(f"{exclusion_mode} mode not supported")


if __name__ == "__main__":

    def asijsda():
        strings = [
            " # aasd # sad ",
            " faojasasd # oiwaos ",
            " okjasifj  oajsidw2 ",
            " 12 329#9213",
        ]
        for s in symbol_filter(strings):
            print(s)

        print(" ")

        for i, s in enumerate(
            symbol_filter(strings, exclusion_mode=FilterModeEnum.exclude_fully)
        ):
            print(i, s)

        print(" ")

        for s in symbol_filter(strings, exclusion_mode=FilterModeEnum.exclude_prefix):
            print(s)

    asijsda()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import torch

from draugr.torch_utilities.tensors import to_tensor

__author__ = "Christian Heider Nielsen"
__all__ = ["TensoriseMixin"]


class TensoriseMixin(object):
    """
  Tensorise attributes at set"""

    device = "cpu"
    dtype = torch.float  # Default values may be monkey patched for other types

    def __setattr__(self, key, value):
        super().__setattr__(key, to_tensor(value, dtype=self.dtype, device=self.device))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Any

import cv2
import numpy

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/12/2019
           """

__all__ = ["cv2_resize"]


def cv2_resize(x: Any, target_size: tuple, interpolation=cv2.INTER_LINEAR) -> Any:
    """

  :param interpolation:
  :param x:
  :param target_size: proper (width, height) shape, no cv craziness
  :return:"""
    if x.shape != target_size:
        x = cv2.resize(x, target_size[::-1], interpolation=interpolation)
    return x


if __name__ == "__main__":
    a = numpy.zeros((50, 50))
    a = cv2_resize(numpy.zeros((100, 100)), a.shape)
    print(a.shape)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/07/2020
           """

import os
from pathlib import Path
from typing import Sequence, Tuple, Union, Iterable

__all__ = ["build_flat_dataset"]


def build_flat_dataset(
    directory: Union[Path, str],
    *,
    validation_percentage: float = 15,
    testing_percentage: float = 0,
    extensions: Iterable = None,
    is_valid_file: callable = None,
) -> dict:
    """

  :param directory:
  :param extensions:
  :param is_valid_file:
  :return:
  """
    if not isinstance(directory, Path):
        directory = Path(directory)

    categories = [d.name for d in directory.iterdir() if d.is_dir()]
    instances = {k: [] for k in categories}
    both_none = extensions is None and is_valid_file is None
    both_something = extensions is not None and is_valid_file is not None
    if both_none or both_something:
        raise ValueError(
            "Both extensions and is_valid_file cannot be None or not None at the same time"
        )

    if extensions is not None:

        def is_valid_file(x: Union[Path, str]) -> bool:
            return str(x) in extensions

    elif is_valid_file is None:
        is_valid_file = lambda a: a is not None

    for target_class in sorted(categories):
        target_dir = directory.expanduser() / target_class
        if not target_dir.is_dir():
            continue
        for root, _, fnames in sorted(os.walk(str(target_dir), followlinks=True)):
            for fname in sorted(fnames):
                path = Path(root) / fname
                if is_valid_file(path):
                    instances[target_class].append(path)
    return instances


if __name__ == "__main__":

    def absa():
        from draugr.visualisation import indent_lines
        from draugr.numpy_utilities.datasets.splitting import Split

        a = build_flat_dataset(Path.home() / "Data" / "mnist_png" / "training")

        for k in a.keys():
            total = (
                len(a[k][Split.Training])
                + len(a[k][Split.Validation])
                + len(a[k][Split.Testing])
            )
            print(f"\n{k}:")
            print(indent_lines(len(a[k][Split.Training]) / total))
            print(indent_lines(len(a[k][Split.Validation]) / total))
            print(indent_lines(len(a[k][Split.Testing]) / total))

    absa()



__all__ = ["DEFAULT_ACCEPTED_FILE_FORMATS", "DEFAULT_ACCEPTED_IMAGE_FORMATS"]

DEFAULT_ACCEPTED_IMAGE_FORMATS = (
    "png",
    "PNG",
    "jpg",
    "JPG",
    "jpeg",
    "JPEG",
    "gif",
    "GIF",
)

DEFAULT_ACCEPTED_FILE_FORMATS = DEFAULT_ACCEPTED_IMAGE_FORMATS



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03/04/2020
           """

import time

import numpy


def test_perf_delete_vs_view():
    samples = 10
    for _ in range(samples):
        a = numpy.zeros((1000, 1000))
        s1 = time.time()
        b = a[1:]
        s2 = time.time()
        c = numpy.delete(a, 0, 0)
        s3 = time.time()

        print(f"view: {s2 - s1}")
        print(f"delete: {s3 - s2}")


def test_perf_delete_vs_view2():
    samples = 10
    for _ in range(samples):
        a = numpy.zeros((1000, 1000))
        s1 = time.time()
        c = numpy.delete(a, 0, 0)
        s2 = time.time()
        b = a[1:]
        s3 = time.time()

        print(f"delete: {s2 - s1}")
        print(f"view: {s3 - s2}")

    # assert b == c



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 16-12-2020
           """

from typing import Iterable, Sequence

__all__ = ["last_dim_truncate", "min_length_truncate_batch", "truncate_to_power_2"]

from draugr.python_utilities.powers import prev_pow_2


def min_length_truncate_batch(batch: Iterable[Sequence]) -> Iterable:
    """

  :param batch:
  :return:
  """
    min_seq_len = min([s.shape[-1] for s in batch])
    return [last_dim_truncate(s, min_seq_len) for s in batch]


def min_length_truncate_batch_2d(batch: Iterable[Sequence]) -> Iterable:
    """

  :param batch:
  :return:
  """
    min_seq_len = min([len(s) for s in batch])
    return [last_dim_truncate(s, min_seq_len) for s in batch]


def last_dim_truncate(sequence: Sequence, min_length: int) -> Sequence:
    """

  :param sequence:
  :param min_length:
  :return:
  """
    return sequence[..., :min_length]


def truncate_to_power_2(signal: Sequence) -> Sequence:
    return last_dim_truncate(signal, prev_pow_2(len(signal)))


if __name__ == "__main__":

    def gasdasa():
        from draugr.torch_utilities import to_tensor

        base = 5
        stair_length = 9
        stair = [to_tensor(range(i + base)) for i in range(stair_length)]

        trunc = [last_dim_truncate(s, base) for s in stair]
        print(to_tensor(trunc))

    def basdiuj():
        import numpy

        asda = numpy.arange(2 ** 5 - 1)
        print(asda, len(asda))
        trunc = truncate_to_power_2(asda)
        print(trunc, len(trunc))

    # gasdasa()
    basdiuj()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 30-12-2020
           """

from draugr.tqdm_utilities import progress_bar


def test_progress_bar():
    for a in progress_bar(range(100), notifications=False):
        pass


def test_asudhweasijdq():
    sas = [f"a{a}" for a in range(5)]

    for a in progress_bar(sas, notifications=False):
        pass


def test_asudhwea213si23jdq():
    def sas(l):
        yield from range(l)

    for a in progress_bar(sas(129), notifications=False):
        pass


def test_asudhweasi23jdq():
    def sas():
        yield from range(100)

    for a in progress_bar(sas(), notifications=False):
        pass


def test_dsad3123():
    for a in progress_bar([2.13, 8921.9123, 923], notifications=False):
        pass


def test_dsad311231223():
    for a in progress_bar(
        (
            2.13,
            8921.9123,
            923,
            821738,
            782173,
            8912738124,
            8471827,
            661262,
            1111,
            2222,
            3333,
            4444,
            5555,
        ),
        notifications=False,
    ):
        pass


def test_ds12sadad311231223():
    for a in progress_bar(
        {
            2.13j,
            8921.9123j,
            923j,
            821738j,
            782173j,
            8912738124j,
            8471827j,
            661262j,
            1111j,
            2222j,
            3333j,
            4444j,
            5555j,
        },
        notifications=False,
    ):
        pass


def test_ds12s23():
    for a in progress_bar(
        [[2.13j], [8921.9123j], [923j], [821738j], [782173j]], notifications=False
    ):
        pass


def test_ds1saijd2s23():
    a1 = range(6)
    b2 = range(6)

    for _ in progress_bar(a1, notifications=False):
        pass

    for _ in progress_bar(b2, notifications=False):
        pass


def test_dict_items():
    from time import sleep

    class exp_v:
        Test_Sets = {v: v for v in range(9)}

    for a in progress_bar(exp_v.Test_Sets.items()):
        sleep(1)


if __name__ == "__main__":
    test_progress_bar()
    test_asudhweasijdq()
    test_asudhweasi23jdq()
    test_dsad3123()
    test_dsad311231223()
    test_ds12sadad311231223()
    test_ds12s23()
    test_asudhwea213si23jdq()
    test_ds1saijd2s23()
    test_dict_items()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pathlib
from contextlib import suppress
from typing import Any, Iterable, Sequence, Union

import PIL
import numpy
import torch
from PIL import Image
from matplotlib import pyplot
from matplotlib.figure import Figure

from draugr import PROJECT_APP_PATH
from draugr.python_utilities import sprint
from draugr.torch_utilities import to_tensor
from draugr.writers import Writer
from draugr.writers.mixins import (
    BarWriterMixin,
    EmbedWriterMixin,
    GraphWriterMixin,
    HistogramWriterMixin,
    ImageWriterMixin,
)
from draugr.writers.mixins.figure_writer_mixin import FigureWriterMixin
from draugr.writers.mixins.instantiation_writer_mixin import InstantiationWriterMixin
from draugr.writers.mixins.line_writer_mixin import LineWriterMixin
from draugr.writers.mixins.precision_recall_writer_mixin import (
    PrecisionRecallCurveWriterMixin,
)
from draugr.writers.mixins.spectrogram_writer_mixin import SpectrogramWriterMixin
from warg import drop_unused_kws, passes_kws_to

with suppress(FutureWarning):
    from torch.utils.tensorboard import SummaryWriter

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""
__all__ = ["TensorBoardPytorchWriter"]


class TensorBoardPytorchWriter(
    Writer,
    ImageWriterMixin,
    GraphWriterMixin,
    HistogramWriterMixin,
    BarWriterMixin,
    LineWriterMixin,
    SpectrogramWriterMixin,
    FigureWriterMixin,
    InstantiationWriterMixin,
    PrecisionRecallCurveWriterMixin,
    EmbedWriterMixin,
):
    """
  Provides a pytorch-tensorboard-implementation writer interface"""

    @passes_kws_to(SummaryWriter.add_embedding)
    def embed(
        self,
        tag: str,
        response: Sequence,
        metadata: Any = None,
        label_img: Any = None,
        step: int = None,
        **kwargs,
    ) -> None:
        """

BORKED!

:param tag:
:param response:
:param metadata:
:param label_img:
:param step:
:param kwargs:
:return:
"""
        try:
            self.writer.add_embedding(
                response,
                metadata=metadata,
                label_img=label_img,
                global_step=step,
                tag=tag,
                **kwargs,
            )
        except Exception as e:
            sprint("Try update tensorflow and/or tensorboard")
            raise e

    @passes_kws_to(Writer.__init__)
    def __init__(
        self,
        path: Union[str, pathlib.Path] = pathlib.Path.cwd() / "Logs",
        summary_writer_kws=None,
        **kwargs,
    ):
        """

:param path:
:param summary_writer_kws:
:param kwargs:
"""
        super().__init__(**kwargs)

        if summary_writer_kws is None:
            summary_writer_kws = {}

        self._log_dir = path
        self._summary_writer_kws = summary_writer_kws

    # @passes_kws_to(SummaryWriter.add_hparams)
    def instance(self, instance: dict, metrics: dict) -> None:
        """

Not finished!

:param instance:
:param metrics:
:return:
"""
        self.writer.add_hparams(instance, metrics)

    @drop_unused_kws
    @passes_kws_to(SummaryWriter.add_pr_curve)
    def precision_recall_curve(
        self,
        tag: str,
        predictions: Iterable,
        truths: Iterable,
        step: int = None,
        **kwargs,
    ) -> None:
        """

    :param tag:
    :param predictions:
    :param truths:
    :param step:
    :param kwargs:
    """
        self.writer.add_pr_curve(
            tag,
            to_tensor(truths, device="cpu"),
            to_tensor(predictions, device="cpu"),
            global_step=step,
            **kwargs,
        )

    @drop_unused_kws
    @passes_kws_to(SummaryWriter.add_figure)
    def figure(self, tag: str, figure: Figure, step: int, **kwargs) -> None:
        """

    :param tag:
    :type tag:
    :param figure:
    :type figure:
    :param step:
    :type step:
    :param kwargs:
    :type kwargs:"""
        self.writer.add_figure(tag, figure, global_step=step, **kwargs)

    @drop_unused_kws
    @passes_kws_to(SummaryWriter.add_figure)
    def spectrogram(
        self,
        tag: str,
        values: list,
        sample_rate: int,
        step: int,
        n_fft: int = 512,
        step_size=128,
        x_labels: Sequence = None,
        y_label: str = "Frequency [Hz]",
        x_label: str = "Time [sec]",
        plot_kws=None,
        **kwargs,
    ) -> None:
        """

        :param sample_rate:
        :param n_fft:
        :param step_size:
    :param tag:
    :type tag:
    :param values:
    :type values:
    :param step:
    :type step:
    :param x_labels:
    :type x_labels:
    :param y_label:
    :type y_label:
    :param x_label:
    :type x_label:
    :param plot_kws:
    :type plot_kws:
    :param kwargs:
    :type kwargs:"""
        if plot_kws is None:
            plot_kws = {}
        fig = pyplot.figure()

        spec = pyplot.specgram(
            values, NFFT=n_fft, Fs=sample_rate, noverlap=step_size, **plot_kws
        )

        """
ind = numpy.arange(len(values))
if x_labels:
pyplot.xticks(ind, labels=x_labels)
else:
pyplot.xticks(ind)

pyplot.xlabel(x_label)
pyplot.ylabel(y_label)
pyplot.title(tag)
"""

        pyplot.colorbar()
        self.writer.add_figure(
            tag, fig, global_step=step, close=True, **kwargs
        )  # TODO: Pull out hardcoded kws to argument list for this method, duplicate kwarg issue possible

    # def embed(self, tag: str, features, metadata, label_img, step: int, **kwargs) -> None:

    @drop_unused_kws
    @passes_kws_to(SummaryWriter.add_figure)
    def bar(
        self,
        tag: str,
        values: list,
        step: int,
        value_error: float = None,
        x_labels: Sequence = None,
        y_label: str = "Probabilities",
        x_label: str = "Distribution",
        **kwargs,
    ) -> None:
        """

    :param x_label:
    :type x_label:
    :param tag:
    :type tag:
    :param values:
    :type values:
    :param step:
    :type step:
    :param value_error:
    :type value_error:
    :param x_labels:
    :type x_labels:
    :param y_label:
    :type y_label:
    :param kwargs:
    :type kwargs:"""
        fig = pyplot.figure()
        ind = numpy.arange(len(values))
        im = pyplot.bar(ind, values, yerr=value_error)
        if x_labels:
            pyplot.xticks(ind, labels=x_labels)
        else:
            pyplot.xticks(ind)

        pyplot.xlabel(x_label)
        pyplot.ylabel(y_label)
        pyplot.title(tag)

        self.writer.add_figure(
            tag, fig, global_step=step, close=True, **kwargs
        )  # TODO: Pull out hardcoded kws to argument list for this method, duplicate kwarg issue possible

    @drop_unused_kws
    @passes_kws_to(SummaryWriter.add_figure)
    def line(
        self,
        tag: str,
        values: list,
        step: int,
        x_labels: Sequence = None,
        y_label: str = "Magnitude",
        x_label: str = "Sequence",
        plot_kws=None,  # Separate as parameters name collisions might occur
        **kwargs,
    ) -> None:
        """

    :param x_label:
    :type x_label:
    :param plot_kws:
    :type plot_kws:
    :param tag:
    :type tag:
    :param values:
    :type values:
    :param step:
    :type step:
    :param x_labels:
    :type x_labels:
    :param y_label:
    :type y_label:
    :param kwargs:
    :type kwargs:"""
        if plot_kws is None:
            plot_kws = {}
        fig = pyplot.figure()
        ind = numpy.arange(len(values))
        im = pyplot.plot(values, **plot_kws)
        if x_labels:
            pyplot.xticks(ind, labels=x_labels)
        else:
            pyplot.xticks(ind)

        pyplot.xlabel(x_label)
        pyplot.ylabel(y_label)
        pyplot.title(tag)

        self.writer.add_figure(
            tag, fig, global_step=step, close=True, **kwargs
        )  # TODO: Pull out hardcoded kws to argument list for this method, duplicate kwarg issue possible

    @drop_unused_kws
    @passes_kws_to(SummaryWriter.add_histogram)
    def histogram(
        self, tag: str, values: list, step: int, bins: Any = "auto", **kwargs
    ) -> None:
        """

    :param tag:
    :type tag:
    :param values:
    :type values:
    :param step:
    :type step:
    :param bins:
    :type bins:
    :param kwargs:
    :type kwargs:"""
        self.writer.add_histogram(tag, values, global_step=step, bins=bins, **kwargs)

    def _scalar(self, tag: str, value: float, step: int) -> None:
        self.writer.add_scalar(tag, value, step)

    @drop_unused_kws
    @passes_kws_to(SummaryWriter.add_graph)
    def graph(
        self,
        model: torch.nn.Module,
        input_to_model: torch.Tensor,
        verbose: bool = False,
    ) -> None:
        """

        :param verbose:
    :param model:
    :type model:
    :param input_to_model:
    :type input_to_model:"""
        self.writer.add_graph(model, input_to_model, verbose)

    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        if hasattr(self, "_writer"):
            self._writer.close()
            delattr(self, "_writer")

    @passes_kws_to(SummaryWriter.add_image)
    def image(
        self,
        tag: str,
        data: Union[numpy.ndarray, torch.Tensor, PIL.Image.Image],
        step,
        *,
        data_formats="NCHW",
        **kwargs,
    ) -> None:
        """

    :param tag:
    :type tag:
    :param data:
    :type data:
    :param step:
    :type step:
    :param data_formats:
    :type data_formats:
    :param kwargs:
    :type kwargs:"""
        self.writer.add_image(tag, data, step, dataformats=data_formats, **kwargs)

    @property
    def writer(self) -> SummaryWriter:
        """

    :return:
    :rtype:"""
        if not hasattr(self, "_writer") or not self._writer:
            self._writer = SummaryWriter(str(self._log_dir), **self._summary_writer_kws)
            if self._verbose:
                print(f"Logging at {self._log_dir}")
        return self._writer

    def _open(self):
        return self


# NAMING AND CASING, SO MANY IMPLEMENTATION OF TENSORBOARD WRITERS, COVER BASES!
TensorboardPytorchWriter = TensorBoardPytorchWriter
TensorBoardPyTorchWriter = TensorBoardPytorchWriter
TensorboardPyTorchWriter = TensorBoardPytorchWriter
PytorchTensorboardWriter = TensorBoardPytorchWriter
PyTorchTensorBoardWriter = TensorBoardPytorchWriter
PyTorchTensorboardWriter = TensorBoardPytorchWriter
TensorboardTorchWriter = TensorBoardPytorchWriter
TensorBoardTorchWriter = TensorBoardPytorchWriter
TorchTensorboardWriter = TensorBoardPytorchWriter
TorchTensorBoardWriter = TensorBoardPytorchWriter

if __name__ == "__main__":

    with TensorBoardPytorchWriter(PROJECT_APP_PATH.user_log / "test") as writer:
        writer.scalar("What", 4)

        from torchvision.utils import make_grid

        for n_iter in range(20):
            dummy_img = torch.rand(32, 3, 64, 64)  # output from network
            if n_iter % 10 == 0:
                x = make_grid(dummy_img, normalize=True, scale_each=True)
                writer.image("ImageGrid", x, n_iter, data_formats="CHW")
                writer.image("Image", dummy_img, n_iter)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-01-2021
           """

__all__ = ["duplicate_columns"]

from typing import List

import pandas
from pandas.core.dtypes.missing import array_equivalent


def duplicate_columns(frame: pandas.DataFrame) -> List[str]:
    groups = frame.columns.to_series().groupby(frame.dtypes).groups
    dups = []

    for t, v in groups.items():

        cs = frame[v].columns
        vs = frame[v]
        lcs = len(cs)

        for i in range(lcs):
            ia = vs.iloc[:, i].values
            for j in range(i + 1, lcs):
                ja = vs.iloc[:, j].values
                if array_equivalent(ia, ja):
                    dups.append(cs[i])
                    break

    return dups



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11/05/2020
           """

import torch

from draugr.torch_utilities import (
    TorchCpuSession,
    TorchCudaSession,
    TorchEvalSession,
    TorchTrainSession,
    global_torch_device,
)


def test_cpu():
    print(
        global_torch_device(override=global_torch_device(cuda_device_preference=True))
    )
    print(global_torch_device())
    with TorchCpuSession():
        print(global_torch_device())
    with TorchCudaSession():
        print(global_torch_device())
    print(
        global_torch_device(override=global_torch_device(cuda_device_preference="cpu"))
    )
    print(global_torch_device())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 07/07/2020
           """

import librosa
import numpy
import torch
from librosa.display import specshow
from matplotlib import pyplot

from draugr import PROJECT_APP_PATH
from draugr.torch_utilities import (
    TensorBoardPytorchWriter,
    constant_init,
    fan_in_init,
    normal_init,
    weight_bias_histograms,
    xavier_init,
)

if __name__ == "__main__":
    NFFT = 256
    STEP_SIZE = NFFT // 2

    DELTA = 0.001
    time_ = numpy.arange(0, 1, DELTA)
    SAMPLING_RATE = int(1 / DELTA)

    SIGNAL = numpy.sin(2 * numpy.pi * 50 * time_) + numpy.sin(
        2 * numpy.pi * 120 * time_
    )

    def module_param_histograms():

        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            input_f = 4
            n_classes = 10
            num_updates = 20

            model = torch.nn.Sequential(
                torch.nn.Linear(input_f, 20),
                torch.nn.ReLU(),
                torch.nn.Linear(20, n_classes),
                torch.nn.LogSoftmax(-1),
            )

            for i in range(num_updates):
                normal_init(
                    model, 0.2 * float((i - num_updates * 0.5) ** 2), 1 / (i + 1)
                )
                weight_bias_histograms(writer, model, step=i, prefix="normal")

                xavier_init(model)
                weight_bias_histograms(writer, model, step=i, prefix="xavier")

                constant_init(model, i)
                weight_bias_histograms(writer, model, step=i, prefix="constant")

                fan_in_init(model)
                weight_bias_histograms(writer, model, step=i, prefix="fan_in")

    def signal_plot():
        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            writer.line("Signal", SIGNAL, step=0)

    def fft_plot():
        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            spectral = numpy.fft.fft(SIGNAL, NFFT)
            writer.line("FFT", spectral, title="Frequency", step=0)

    def spectral_plot():
        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            writer.spectrogram(
                "STFT", SIGNAL, int(1 / DELTA), step=0, n_fft=NFFT, step_size=STEP_SIZE
            )

    def spectral_plot_scipy():
        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            writer.spectrogram(
                "STFT_Scipy",
                SIGNAL,
                int(1 / DELTA),
                step=0,
                n_fft=NFFT,
                step_size=STEP_SIZE,
            )

    def cepstral_plot():
        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            fig = pyplot.figure()
            stft = librosa.core.stft(SIGNAL, n_fft=NFFT, hop_length=STEP_SIZE)
            specshow(stft, sr=SAMPLING_RATE, x_axis="time")
            pyplot.colorbar()
            writer.figure("STFT_Rosa", fig, step=0)

    def mel_cepstral_plot():
        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            fig = pyplot.figure()
            mfccs = librosa.feature.mfcc(
                SIGNAL, sr=SAMPLING_RATE, n_mfcc=20, n_fft=NFFT, hop_length=STEP_SIZE
            )
            specshow(mfccs, sr=SAMPLING_RATE, x_axis="time")
            pyplot.colorbar()
            writer.figure("MFCC_Rosa", fig, step=0)

    module_param_histograms()
    signal_plot()
    fft_plot()
    spectral_plot()
    spectral_plot_scipy()
    cepstral_plot()
    mel_cepstral_plot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

__doc__ = r"""
"""

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .image_data import *
from .matplotlib_utilities import *
from .metric_overview_plot import *
from .pillow_utilities import *
from .signal_data import *
from .figure_sessions import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# import numba
import numpy

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """

__all__ = ["hwc_to_chw", "chw_to_hwc", "uint_hwc_to_chw_float", "float_chw_to_hwc_uint"]


# @numba.njit()
def hwc_to_chw(inp: numpy.ndarray) -> numpy.ndarray:
    """

  :param inp:
  :type inp:
  :return:
  :rtype:"""
    return inp.transpose((2, 0, 1))


# @numba.njit()
def chw_to_hwc(inp: numpy.ndarray) -> numpy.ndarray:
    """

  :param inp:
  :type inp:
  :return:
  :rtype:"""
    return inp.transpose((1, 2, 0))


# @numba.njit()
def uint_hwc_to_chw_float(
    inp: numpy.ndarray, *, normalise: bool = True
) -> numpy.ndarray:
    """

  :param inp:
  :type inp:
  :param normalise:
  :type normalise:
  :return:
  :rtype:"""
    if normalise:
        inp /= 255.0
        inp = numpy.clip(inp, 0, 1)
    return hwc_to_chw(inp)


# @numba.njit()
def float_chw_to_hwc_uint(
    inp: numpy.ndarray, *, unnormalise: bool = True
) -> numpy.ndarray:
    """

  :param inp:
  :type inp:
  :param unnormalise:
  :type unnormalise:
  :return:
  :rtype:"""
    inp = chw_to_hwc(inp)
    if unnormalise:
        inp *= 255.0
        inp = numpy.clip(inp, 0, 255)
    return inp.astype(numpy.uint8)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/07/2020
           """

import random
from pathlib import Path
from typing import Iterable, Tuple

from torch.utils.data import DataLoader
from torchvision.datasets import VisionDataset  # TODO: Do not need to be images

__all__ = ["DictDatasetFolder", "SplitDictDatasetFolder"]

from torchvision.datasets.folder import has_file_allowed_extension

from draugr.numpy_utilities import (
    Split,
    build_flat_dataset,
    build_shallow_categorical_dataset,
    select_split,
)


class SplitDictDatasetFolder(VisionDataset):
    """A generic data loader where the samples are arranged in this way: ::

root/class_x/xxx.ext
root/class_x/xxy.ext
root/class_x/xxz.ext

root/class_y/123.ext
root/class_y/nsdf3.ext
root/class_y/asd932_.ext

Args:
root (string): Root directory path.
loader (callable): A function to load a sample given its path.
extensions (tuple[string]): A list of allowed extensions.
    both extensions and is_valid_file should not be passed.
transform (callable, optional): A function/transform that takes in
    a sample and returns a transformed version.
    E.g, ``transforms.RandomCrop`` for images.
target_transform (callable, optional): A function/transform that takes
    in the target and transforms it.
is_valid_file (callable, optional): A function that takes path of a file
    and check if the file is a valid file (used to check of corrupt files)
    both extensions and is_valid_file should not be passed.

Attributes:
_categories (list): List of the class names sorted alphabetically.
_data_categories (list): List of (sample path, class_index) tuples"""

    def __init__(
        self,
        root: Path,
        loader: DataLoader,
        extensions: Iterable = None,
        transform: callable = None,
        target_transform: callable = None,
        split: Split = Split.Training,
        valid_percentage: float = 15,
        test_percentage: float = 0,
        is_valid_file: callable = has_file_allowed_extension,
    ):
        super().__init__(
            str(root), transform=transform, target_transform=target_transform
        )
        # TODO: merge Split and non split common in a base class
        self._data_categories = select_split(
            build_shallow_categorical_dataset(
                self.root,
                extensions=extensions,
                testing_percentage=test_percentage,
                validation_percentage=valid_percentage,
            ),
            split,
        )

        if len(self._data_categories) == 0:
            msg = f"Found 0 categories in sub-folders of: {self.root}\n"
            if extensions is not None:
                msg += f"Supported extensions are: {','.join(extensions)}"
            raise RuntimeError(msg)

        self.loader = loader
        self.extensions = extensions

        self.category_sizes = {k: len(v) for k, v in self._data_categories.items()}
        for cat, cl in self.category_sizes.items():
            if cl == 0:
                print(f"Warning category {cat} has {cl} samples")
        self.category_names = (*self.category_sizes.keys(),)

    def __getitem__(self, index) -> Tuple:
        """

Non-pure implementation! Index maybe not map to the same item as target randomly sampled

Args:
index (int): Index

Returns:
tuple: (sample, target) where target is class_index of the target class."""
        target = random.choice(self.category_names)
        return self.sample(target, index)

    def sample(self, target, index) -> Tuple:
        sample = self.loader(
            self._data_categories[target][index % self.category_sizes[target]]
        )
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return sample, target

    def __len__(self):
        return sum(list(self.category_sizes.values()))


class DictDatasetFolder(VisionDataset):
    """A generic data loader where the samples are arranged in this way: ::

root/class_x/xxx.ext
root/class_x/xxy.ext
root/class_x/xxz.ext

root/class_y/123.ext
root/class_y/nsdf3.ext
root/class_y/asd932_.ext

Args:
root (string): Root directory path.
loader (callable): A function to load a sample given its path.
extensions (tuple[string]): A list of allowed extensions.
    both extensions and is_valid_file should not be passed.
transform (callable, optional): A function/transform that takes in
    a sample and returns a transformed version.
    E.g, ``transforms.RandomCrop`` for images.
target_transform (callable, optional): A function/transform that takes
    in the target and transforms it.
is_valid_file (callable, optional): A function that takes path of a file
    and check if the file is a valid file (used to check of corrupt files)
    both extensions and is_valid_file should not be passed.

Attributes:
_categories (list): List of the class names sorted alphabetically.
_data (list): List of (sample path, class_index) tuples"""

    def __init__(
        self,
        root: Path,
        loader: DataLoader,
        extensions: Iterable = None,
        transform: callable = None,
        target_transform: callable = None,
        is_valid_file: callable = has_file_allowed_extension,
    ):
        super().__init__(
            str(root), transform=transform, target_transform=target_transform
        )
        # TODO: merge Split and non split common in a base class
        self._data = build_flat_dataset(
            self.root, extensions=extensions, is_valid_file=is_valid_file
        )

        if len(self._data) == 0:
            msg = f"Found 0 files in sub-folders of: {self.root}\n"
            if extensions is not None:
                msg += f"Supported extensions are: {','.join(extensions)}"
            raise RuntimeError(msg)

        self.loader = loader
        self.extensions = extensions

        self.category_sizes = {k: len(v) for k, v in self._data.items()}
        self.category_names = (*self.category_sizes.keys(),)

    def __getitem__(self, index) -> Tuple:
        """

Non-pure implementation! Index maybe not map to the same item as target randomly sampled

Args:
index (int): Index

Returns:
tuple: (sample, target) where target is class_index of the target class."""
        target = random.choice(self.category_names)
        return self.sample(target, index)

    def sample(self, target, index) -> Tuple:
        sample = self.loader(self._data[target][index % self.category_sizes[target]])
        if self.transform is not None:
            sample = self.transform(sample)
        if self.target_transform is not None:
            target = self.target_transform(target)
        return sample, target

    def __len__(self):
        return sum(list(self.category_sizes.values()))


if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02/03/2020
           """

__all__ = ["NoData", "IncompatiblePackageVersions"]


class NoData(Exception):
    """"""

    def __init__(self, msg: str = "No Data Available"):
        Exception.__init__(self, msg)


class IncompatiblePackageVersions(Exception):
    """"""

    def __init__(self, *packages, **versions):
        str_o = ", "
        str_l = []

        for p in packages:
            if isinstance(p, str):
                s = f"({p},"
                if p in versions:
                    s += f"{versions[p]}"
                else:
                    s += f"NotSpecified"
                s += ")"
                str_l.append(s)
            else:
                str_l.append(f"{p.__name__, p.__version__}")

        Exception.__init__(self, f"Packages {str_o.join(str_l)} are not compatible")


if __name__ == "__main__":

    def main():
        raise IncompatiblePackageVersions(
            "numpy", "scipy", "Something", numpy="0.0.1", scipy="0.0.2"
        )

    def main2():
        import numpy
        import scipy

        raise IncompatiblePackageVersions(numpy, scipy)

    main2()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

import math
from typing import Tuple, Union

from draugr import replicate

__all__ = ["conv2d_padding_size", "conv2d_hw_shape"]


def conv2d_hw_shape(
    h_w: Union[int, Tuple[int, int]],
    kernel_size: Union[int, Tuple[int, int]] = 1,
    stride: Union[int, Tuple[int, int]] = 1,
    pad: Union[int, Tuple[int, int], Tuple[int, int, int, int]] = 0,
    dilation: Union[int, Tuple[int, int]] = 1,
) -> Tuple[int, int]:
    """

  :param h_w:
  :type h_w:
  :param kernel_size:
  :type kernel_size:
  :param stride:
  :type stride:
  :param pad:
  :type pad:
  :param dilation:
  :type dilation:
  :return:
  :rtype:"""
    (h_w, kernel_size, stride, pad, dilation) = (
        replicate(h_w),
        replicate(kernel_size),
        replicate(stride),
        replicate(pad),
        replicate(dilation),
    )
    pad = (replicate(pad[0]), replicate(pad[1]))

    h = math.floor(
        (h_w[0] + sum(pad[0]) - dilation[0] * (kernel_size[0] - 1) - 1) / stride[0] + 1
    )
    w = math.floor(
        (h_w[1] + sum(pad[1]) - dilation[1] * (kernel_size[1] - 1) - 1) / stride[1] + 1
    )

    return h, w


def conv2d_padding_size(
    h_w_in: Union[int, Tuple[int, int]],
    h_w_out: Union[int, Tuple[int, int]],
    kernel_size: int = 1,
    stride: int = 1,
    dilation: int = 1,
) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    """

  :param h_w_in:
  :type h_w_in:
  :param h_w_out:
  :type h_w_out:
  :param kernel_size:
  :type kernel_size:
  :param stride:
  :type stride:
  :param dilation:
  :type dilation:
  :return:
  :rtype:"""
    (h_w_in, h_w_out, kernel_size, stride, dilation) = (
        replicate(h_w_in),
        replicate(h_w_out),
        replicate(kernel_size),
        replicate(stride),
        replicate(dilation),
    )

    p_h = (
        (h_w_out[0] - 1) * stride[0]
        - h_w_in[0]
        + dilation[0] * (kernel_size[0] - 1)
        + 1
    )
    p_w = (
        (h_w_out[1] - 1) * stride[1]
        - h_w_in[1]
        + dilation[1] * (kernel_size[1] - 1)
        + 1
    )

    return (
        (math.floor(p_h / 2), math.ceil(p_h / 2)),
        (math.floor(p_w / 2), math.ceil(p_w / 2)),
    )


if __name__ == "__main__":
    print(conv2d_hw_shape(105, (0, 1, 2, 3)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pytest

from draugr import PROJECT_APP_PATH
from draugr.writers import CSVWriter

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
           """


@pytest.mark.parametrize(
    ["tag", "val", "step"],
    (("signal", 0, 0), ("signal", 20, 1), ("signal", -1, 6)),
    ids=["signal_first", "signal_second", "signal_sixth"],
)
def test_valid_scalars(tag, val, step):
    with CSVWriter(path=PROJECT_APP_PATH.user_log / "csv_writer") as w:
        w.scalar(tag, val, step)


@pytest.mark.parametrize(
    ["tag", "val", "step"],
    (("signal", "", 0), ("signal", None, 1), ("signal", object(), 6)),
    ids=["str_scalar", "None_scalar", "object_scalar"],
)
def test_invalid_val_type_scalars(tag, val, step):
    try:
        with CSVWriter(path=PROJECT_APP_PATH.user_log / "csv_writer") as w:
            w.scalar(tag, val, step)
        assert False
    except Exception as e:
        assert True


@pytest.mark.parametrize(
    ["tag", "val", "step"],
    ((1, 0, 0), (None, 20, 1), (object(), -1, 6)),
    ids=["numeral_tag", "None_tag", "object_tag"],
)
def test_invalid_tag_scalars(tag, val, step):
    try:
        with CSVWriter(path=PROJECT_APP_PATH.user_log / "csv_writer") as w:
            w.scalar(tag, val, step)
        assert False
    except Exception as e:
        print(e)
        assert True


@pytest.mark.parametrize(
    ["tag", "val", "step"],
    (("signal", 0, ""), ("signal", 20, None), ("tag1", -0, object())),
    ids=["str_step", "None_step", "object_step"],
)
def test_invalid_step_type_scalars(tag, val, step):
    try:
        with CSVWriter(path=PROJECT_APP_PATH.user_log / "csv_writer") as w:
            w.scalar(tag, val, step)
        assert False
    except Exception as e:
        print(e)
        assert True



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ = this_init_file.read()

from .system import *
from .datasets import *
from .distributions import *
from .generators import *
from .images import *
from .operations import *
from .optimisation import *
from .persistence import *
from .tensors import *
from .writers import *
from .sessions import *
from .evaluation import *

if __name__ == "__main__":
    print(__doc__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 08-12-2020
           """

from itertools import tee
from typing import Any, Generator, Iterable

import tqdm
from notus.notification import JobNotificationSession

from warg import drop_unused_kws, passes_kws_to

__all__ = ["progress_bar"]


@drop_unused_kws
@passes_kws_to(tqdm.tqdm)
def progress_bar(
    iterable: Iterable,
    description: str = None,
    *,
    leave: bool = False,
    notifications: bool = False,
    total: int = None,
    auto_total_generator: bool = True,
    auto_describe_iterator: bool = True,  # DOES NOT WORK IS THIS FUNCTION IS ALIAS does not match!
    alias="progress_bar",
    **kwargs
) -> Any:
    if description is None and auto_describe_iterator:
        from warg import get_first_arg_name

        description = get_first_arg_name(alias)

    if total is None and isinstance(iterable, Generator) and auto_total_generator:
        iterable, ic = tee(iterable, 2)
        total = len(list(ic))
    if notifications:
        with JobNotificationSession(description):
            yield from tqdm.tqdm(
                iterable, description, leave=leave, total=total, **kwargs
            )
        return
    yield from tqdm.tqdm(iterable, description, leave=leave, total=total, **kwargs)


if __name__ == "__main__":

    def dsad3123():
        from time import sleep

        for a in progress_bar([2.13, 8921.9123, 923], notifications=False):
            sleep(1)

    def asd21sa():
        from time import sleep

        pb = progress_bar  # Aliased!

        for a in pb([2.13, 8921.9123, 923], notifications=False):
            sleep(1)

    def dict_items():
        from time import sleep

        class exp_v:
            Test_Sets = {v: v for v in range(9)}

        for a in progress_bar(exp_v.Test_Sets.items()):
            sleep(1)

    # dsad3123()
    # asd21sa()
    dict_items()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14-11-2020
           """
__all__ = ["quick_to_pil_image"]

import numpy
import torch
from PIL import Image


def quick_to_pil_image(inp: torch.Tensor, mode="RGB") -> Image.Image:
    """

LOTS OF ASSUMPTIONS!

:param inp:
:return:
"""

    return Image.fromarray(inp.cpu().numpy().astype(numpy.uint8), mode)


if __name__ == "__main__":

    def asd2():

        import cv2
        import torch
        from PIL import Image
        from tqdm import tqdm

        from draugr.opencv_utilities import frame_generator
        from draugr.torch_utilities import global_torch_device, to_tensor_generator

        with torch.no_grad():
            for image in tqdm(
                to_tensor_generator(
                    frame_generator(cv2.VideoCapture(0)), device=global_torch_device(),
                )
            ):
                cv2.namedWindow("window_name", cv2.WINDOW_NORMAL)
                cv2.imshow("window_name", numpy.array(quick_to_pil_image(image)))

                if cv2.waitKey(1) == 27:
                    break  # esc to quit

    asd2()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 10/07/2020
           """

import torch

from draugr import PROJECT_APP_PATH
from draugr.torch_utilities.writers.tensorboard import TensorBoardPytorchWriter
from draugr.writers import HistogramWriterMixin

__all__ = ["weight_bias_histograms"]

from warg import passes_kws_to


@passes_kws_to(TensorBoardPytorchWriter.histogram)
def weight_bias_histograms(
    writer: HistogramWriterMixin,
    model: torch.nn.Module,
    *,
    prefix: str = "",
    step: int = 0,
    recurse: bool = True,
    **kwargs
) -> None:
    """

  :param recurse:
  :param writer:
  :type writer:
  :param model:
  :type model:
  :param prefix:
  :type prefix:
  :param step:
  :type step:
  :param kwargs:
  :type kwargs:"""
    for name, param in model.named_parameters(prefix=prefix, recurse=recurse):
        writer.histogram(name, param.clone().cpu().data.numpy(), step, **kwargs)


if __name__ == "__main__":

    def a():
        with TensorBoardPytorchWriter(
            PROJECT_APP_PATH.user_log / "Tests" / "Writers"
        ) as writer:
            input_f = 4
            n_classes = 10

            model = torch.nn.Sequential(
                torch.nn.Linear(input_f, 20),
                torch.nn.ReLU(),
                torch.nn.Linear(20, n_classes),
                torch.nn.LogSoftmax(-1),
            )
            weight_bias_histograms(writer, model)

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18/03/2020
           """

import datetime

__all__ = ["default_datetime_repr"]

DEFAULT_REPRESENTATION = "%Y-%m-%d_%H:%M:%S.%f"


def default_datetime_repr(
    date: datetime, str_format: str = DEFAULT_REPRESENTATION
) -> str:
    """

  :param date:
  :type date:
  :param str_format:
  :type str_format:
  :return:
  :rtype:"""
    return date.strftime(str_format)


def now_repr() -> str:
    return default_datetime_repr(datetime.datetime.now())


if __name__ == "__main__":
    print(now_repr())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["EmbedWriterMixin"]

from typing import Any, Sequence


class EmbedWriterMixin(ABC):
    """
Writer mixin that provides an interface for 'writing' embeds/projections(2d,3d) for interactive visualisation
"""

    @abstractmethod
    def embed(
        self,
        tag: str,
        response: Sequence,
        metadata: Any = None,
        label_img: Any = None,  # raster grid / image / numpy.array
        step: int = None,
        **kwargs
    ) -> None:
        """
    eg. visualising for projections in lower dimensional space

:param tag:
:type tag:
:param step:
:type step:
:param kwargs:
:type kwargs:
"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

import statistics

from draugr.metrics.metric_aggregator import MetricAggregator

MEASURES = statistics.__all__[1:]

__all__ = ["MetricCollection"]


class MetricCollection(dict):
    """"""

    def __init__(
        self,
        metrics=("signal", "length"),
        measures=MEASURES,
        keep_measure_history=True,
        use_disk_cache=True,
    ):
        super().__init__()
        self._metrics = {}
        self._measures = measures
        self._keep_measure_history = keep_measure_history
        self._use_disk_cache = use_disk_cache

        for metric in metrics:
            self._metrics[metric] = MetricAggregator(
                measures=self._measures,
                keep_measure_history=self._keep_measure_history,
                use_disk_cache=self._use_disk_cache,
            )

    def add_metric(self, name):
        """

    :param name:
    :type name:"""
        self._metrics[name] = MetricAggregator(
            measures=self._measures, keep_measure_history=self._keep_measure_history
        )

    def append(self, *args, **kwargs):
        """

    :param args:
    :type args:
    :param kwargs:
    :type kwargs:"""
        for (arg, (k, v)) in zip(args, self._metrics.items()):
            self._metrics[k].append(arg)

        for (k, v) in kwargs:
            self._metrics[k].append(v)

    def remove_metric(self, name):
        """

    :param name:
    :type name:"""
        del self._metrics[name]

    def __len__(self):
        return len(self._metrics)

    @property
    def metrics(self):
        """

    :return:
    :rtype:"""
        return self._metrics

    def __getattr__(self, name):
        return self.__getitem__(name)

    def __repr__(self):
        return f"<StatisticCollection> {self._metrics} </StatisticCollection>"

    def __str__(self):
        return self.__repr__()

    def __iter__(self):
        return self.metrics

    def __getitem__(self, name):
        if name in self._metrics:
            return self._metrics[name]
        else:
            # return self.add_metric(name)
            raise AttributeError

    # def __setitem__(self, key, value):
    #     if key in self._metrics:
    #         if self._keep_measure_history:
    #             self._metrics[key].append(value)
    #         else:
    #             self._metrics[key] = value
    #     else:
    #         self.add_metric(key)
    #         self.append({key:value})

    def keys(self):
        """

    :return:
    :rtype:"""
        return self.metrics.keys()

    def __contains__(self, item):
        return item in self.metrics

    def items(self):
        """

    :return:
    :rtype:"""
        return self.metrics.items()

    def save(self, **kwargs):
        """

    :param kwargs:
    :type kwargs:"""
        for key, value in self._metrics.items():
            value.save(stat_name=key, **kwargs)


if __name__ == "__main__":
    stats = MetricCollection(keep_measure_history=False)
    stats2 = MetricCollection(keep_measure_history=True)

    for i in range(10):
        stats.signal.append(i)
        stats2.signal.append(i)

    print(stats)
    print(stats.signal)
    print(stats.length)
    print(stats.length.measures)
    print(stats.signal.measures)
    print(stats.signal.variance)
    print(stats.signal.calc_moving_average())
    print(stats.signal.max)
    print(stats.signal.min)
    print("\n")
    print(stats2)
    print(stats2.signal.min)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

import math
from typing import Tuple, Union

__all__ = ["transp_conv2d_output_shape", "transp_conv2d_padding_sizes"]

from draugr import replicate


def transp_conv2d_output_shape(
    h_w: Union[int, Tuple[int, int]],
    kernel_size: Union[int, Tuple[int, int]] = 1,
    stride: Union[int, Tuple[int, int]] = 1,
    pad: Union[int, Tuple[int, int]] = 0,
    dilation: Union[int, Tuple[int, int]] = 1,
    out_pad: Union[int, Tuple[int, int]] = 0,
) -> Tuple[int, int]:
    (h_w, kernel_size, stride, pad, dilation, out_pad) = (
        replicate(h_w),
        replicate(kernel_size),
        replicate(stride),
        replicate(pad),
        replicate(dilation),
        replicate(out_pad),
    )
    pad = (replicate(pad[0]), replicate(pad[1]))

    h = (
        (h_w[0] - 1) * stride[0]
        - sum(pad[0])
        + dilation[0] * (kernel_size[0] - 1)
        + out_pad[0]
        + 1
    )
    w = (
        (h_w[1] - 1) * stride[1]
        - sum(pad[1])
        + dilation[1] * (kernel_size[1] - 1)
        + out_pad[1]
        + 1
    )

    return h, w


def transp_conv2d_padding_sizes(
    h_w_in: Union[int, Tuple[int, int]],
    h_w_out: Union[int, Tuple[int, int]],
    kernel_size: Union[int, Tuple[int, int]] = 1,
    stride: Union[int, Tuple[int, int]] = 1,
    dilation: Union[int, Tuple[int, int]] = 1,
    out_pad: Union[int, Tuple[int, int]] = 0,
) -> Tuple[Tuple[int, int], Tuple[int, int]]:
    (h_w_in, h_w_out, kernel_size, stride, dilation, out_pad) = (
        replicate(h_w_in),
        replicate(h_w_out),
        replicate(kernel_size),
        replicate(stride),
        replicate(dilation),
        replicate(out_pad),
    )

    p_h = (
        -(
            h_w_out[0]
            - 1
            - out_pad[0]
            - dilation[0] * (kernel_size[0] - 1)
            - (h_w_in[0] - 1) * stride[0]
        )
        / 2
    )
    p_w = (
        -(
            h_w_out[1]
            - 1
            - out_pad[1]
            - dilation[1] * (kernel_size[1] - 1)
            - (h_w_in[1] - 1) * stride[1]
        )
        / 2
    )

    return (
        (math.floor(p_h / 2), math.ceil(p_h / 2)),
        (math.floor(p_w / 2), math.ceil(p_w / 2)),
    )


if __name__ == "__main__":
    print(transp_conv2d_output_shape(105, 10))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/03/2020
           """


from draugr.python_utilities.debug import evaluate_context
from warg import kws_sink, prod, sink


def test_a():
    print(evaluate_context(kws_sink, "str"))
    print(evaluate_context(kws_sink, 2))
    print(evaluate_context(kws_sink, 2.2))

    print(evaluate_context(prod, (2, 2)))

    print(evaluate_context(prod, (2.2, 2.2)))

    print(evaluate_context(prod, (2, 2.2)))

    print(evaluate_context(prod, (2.2, 2)))

    print(evaluate_context(sink, (2, 2), face=(2.2, 2)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import time

import numpy
import pytest
import torch
from draugr import batched_recycle
from draugr.torch_utilities import to_tensor_generator

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/10/2019
           """


@pytest.mark.skip
def test_d1():
    channels_in = 3
    channels_out = 3

    samples = 4
    device = "cuda"
    batches = 10
    batch_size = 32
    data_shape = (batches * batch_size, channels_in, 512, 512)

    model = torch.nn.Sequential(
        torch.nn.Conv2d(channels_in, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
        torch.nn.Conv2d(channels_out, channels_out, (3, 3)),
        torch.nn.ReLU(),
    ).to("cuda")

    for _ in range(samples):
        s1 = time.time()
        for _, a in zip(
            range(batches),
            to_tensor_generator(
                batched_recycle(numpy.random.sample(data_shape), batch_size),
                device=device,
                preload_next=False,
            ),
        ):

            model(a)

        s2 = time.time()
        for _, a in zip(
            range(batches),
            to_tensor_generator(
                batched_recycle(numpy.random.sample(data_shape), batch_size),
                device=device,
            ),
        ):
            model(a)

        s3 = time.time()

        print(s2 - s1)
        print(s3 - s2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/03/2020
           """

from itertools import cycle
from typing import Iterable


def busy_indicator(
    *,
    stream: callable = print,
    indicator_interval: int = 1,
    phases: Iterable[str] = ("◑", "◒", "◐", "◓"),
) -> Iterable:
    """
  You can choose arbitrary phases like ['|','/','-','\\']

  :param stream:
  :type stream:
  :param indicator_interval:
  :type indicator_interval:
  :param phases:
  :type phases:
  :return:
  :rtype:"""

    phases = cycle(phases)
    i = 0
    while True:
        if i % indicator_interval == 0:
            stream(f"{next(phases)}", end="\r", flush=True)
        yield i
        i += 1


if __name__ == "__main__":
    import time

    for i in busy_indicator(indicator_interval=10):
        time.sleep(0.1)
        if i > 100:
            break



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 06-01-2021
           """

from scipy.signal import welch, spectrogram
import numpy
from matplotlib import pyplot
from typing import Sequence
from draugr.python_utilities.powers import next_pow_2
from warg import Number

__all__ = ["spectral_plot", "ltass_plot", "spectrum_plot", "fft_plot"]


def spectral_plot(
    time: Sequence,
    frequencies: Sequence,
    fxt: numpy.ndarray,
    fig_size: Sequence = (4.8, 2.4),
) -> pyplot.Figure:
    """
  return new figure
  """
    assert fxt.shape == (*frequencies.shape, *time.shape)
    f, ax = pyplot.subplots(figsize=fig_size)
    ax.pcolormesh(time, frequencies / 1000, 10 * numpy.log10(fxt), cmap="viridis")
    ax.set_ylabel("Frequency [kHz]")
    ax.set_xlabel("Time [s]")
    return f


def ltass_plot(signal: Sequence, sampling_rate: int, label="#") -> pyplot.Figure:
    """

  """
    n_per_seg = next_pow_2(
        sampling_rate * (20 / 1000)
    )  # 20 ms, next_pow_2 per seg == n_fft

    f, spectrum = welch(
        signal, sampling_rate, window="hanning", nperseg=n_per_seg, scaling="spectrum"
    )  # Average Power spectrum of signal.
    pyplot.semilogy(f, numpy.sqrt(spectrum), label=label)
    pyplot.xlabel("frequency [Hz]")
    pyplot.ylabel("Linear spectrum [V RMS]")
    return pyplot.gcf()


def spectrum_plot(
    signal: Sequence, sampling_rate: int, window_length_ms: Number = (20 / 1000)
) -> pyplot.Figure:
    """

  """
    n_per_seg = next_pow_2(
        sampling_rate * window_length_ms
    )  # 20 ms, next_pow_2 per seg == n_fft
    f, t, fxt = spectrogram(
        signal,
        fs=sampling_rate,
        window="hanning",
        nperseg=n_per_seg,
        scaling="spectrum",
    )

    return spectral_plot(t, f, fxt)


def fft_plot(signal: Sequence, *, line_width: float = 0.2) -> pyplot.Figure:
    """

  """
    n_per_seg = next_pow_2(len(signal))
    spectrum = numpy.fft.fft(signal, n_per_seg)[: n_per_seg // 2]
    frequencies = numpy.fft.fftfreq(n_per_seg)[: n_per_seg // 2]
    pyplot.plot(frequencies, spectrum.real, label="Amplitude", linewidth=line_width)
    pyplot.plot(frequencies, spectrum.imag, label="Phase", linewidth=line_width)
    pyplot.xlabel("Frequency [kHz]")
    pyplot.legend()
    return pyplot.gcf()


if __name__ == "__main__":

    def asdijaisd():

        sr = 1000
        t = numpy.arange(sr * 4) / sr
        # noise = numpy.random.rand(sr * 2) * 0.001
        signal = numpy.sin(200 * 2 * numpy.pi * t)  # + noise

        fft_plot(signal[:256])
        pyplot.show()

        spectrum_plot(signal, sr)
        pyplot.show()

        ltass_plot(signal, sr)
        pyplot.show()

    asdijaisd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11-12-2020
           """

import os
import random

__all__ = ["seed_stack"]


def seed_stack(s: int = 23) -> None:
    from draugr.torch_utilities import torch_seed

    python_seed(s)
    numpy_seed(s)
    torch_seed(s)


def python_seed(s: int = 2318) -> None:
    random.seed(s)
    os.environ["PYTHONHASHSEED"] = str(s)


def numpy_seed(s: int = 78213) -> None:
    import numpy

    numpy.random.seed(s)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25/03/2020
           """

import collections
import hashlib
import re
import sys
from enum import Enum
from pathlib import Path
from typing import Dict, Iterable, OrderedDict, Sequence, Any

import numpy

__all__ = ["Split", "SplitIndexer", "train_valid_test_split", "select_split"]


class Split(Enum):
    """
Split Enum class for selecting splits

"""

    Training = "training"
    Validation = "validation"
    Testing = "testing"


class SplitIndexer:
    """

"""

    default_split_names = {i: i.value for i in Split}

    def __init__(
        self,
        dataset_length: int,
        training: float = 0.7,
        validation: float = 0.2,
        testing: float = 0.1,
    ):
        self.total_num = dataset_length
        splits = numpy.array([training, validation, testing])
        self.normalised_split = splits / sum(splits)
        (
            self.training_percentage,
            self.validation_percentage,
            self.testing_percentage,
        ) = self.normalised_split
        self.training_num, self.validation_num, self.testing_num = self.unnormalised(
            dataset_length
        )

    def shuffled_indices(self) -> dict:
        split_indices = numpy.random.permutation(self.total_num).tolist()

        return {
            Split.Training: self.select_train_indices(split_indices),
            Split.Validation: self.select_validation_indices(split_indices),
            Split.Testing: self.select_testing_indices(split_indices),
        }

    def select_train_indices(self, ind: Sequence) -> Sequence:
        return ind[: self.training_num]

    def select_validation_indices(self, ind: Sequence) -> Sequence:
        if self.validation_num:
            if self.testing_num:
                return ind[self.training_num : -self.testing_num]
            return ind[self.training_num :]
        return []

    def select_testing_indices(self, ind: Sequence) -> Sequence:
        if self.testing_num:
            return ind[-self.testing_num :]
        return []

    def unnormalised(self, num: int, floored: bool = True) -> numpy.ndarray:
        """

:param num:
:type num:
:param floored:
:type floored:
:return:
:rtype:"""
        unnorm = self.normalised_split * num
        if floored:
            unnorm = numpy.floor(unnorm)
        return unnorm.astype(int)

    def __repr__(self) -> str:
        return str(
            {k: n for k, n in zip(self.default_split_names, self.normalised_split)}
        )

    def select_shuffled_split_indices(self, split: Split, seed: int = 0) -> Sequence:
        numpy.random.seed(seed)
        split_indices = numpy.random.permutation(self.total_num).tolist()

        if split == Split.Training:
            return self.select_train_indices(split_indices)
        elif split == Split.Validation:
            return self.select_validation_indices(split_indices)
        elif split == Split.Testing:
            return self.select_testing_indices(split_indices)
        elif split is None:
            return split_indices
        raise NotImplementedError


def train_valid_test_split(
    categories: Dict[str, Iterable[Path]],
    *,
    validation_percentage: float = 15,  # TODO: ACCEPT AND SQUEEZE ZERO-HUNDRED TO ZERO-ONE range!
    testing_percentage: float = 0,
    verbose: bool = False,
) -> OrderedDict:
    """
Magic hashing

:param verbose:
:type verbose:
:param categories:
:param testing_percentage:
:param validation_percentage:
:return:"""
    result = collections.OrderedDict()

    if verbose:
        print(categories)

    for c, vs in categories.items():
        training_images = []
        testing_images = []
        validation_images = []

        for file_name in vs:
            b_rep = bytes(re.sub(r"_nohash_.*$", "", f"{c}{file_name.name}"), "utf8")
            percentage_hash = (
                int(hashlib.sha1(b_rep).hexdigest(), 16) % (sys.maxsize + 1)
            ) * (100.0 / sys.maxsize)
            if percentage_hash < validation_percentage + testing_percentage:
                if percentage_hash < testing_percentage:
                    testing_images.append(file_name)
                else:
                    validation_images.append(file_name)
            else:
                training_images.append(file_name)

        result[c] = {
            Split.Training: training_images,
            Split.Validation: validation_images,
            Split.Testing: testing_images,
        }

    return result


def select_split(
    data_cat_split: Dict[Any, Dict[Split, Sequence]],
    split: Split,
    verbose: bool = False,
) -> Dict[Any, Sequence]:
    """

:param verbose:
:type verbose:
:param data_cat_split:
:type data_cat_split:
:param split:
:type split:
:return:
:rtype:"""
    data = {k: [] for k in data_cat_split.keys()}
    if verbose:
        print(data_cat_split)
    for k, v in data_cat_split.items():
        if verbose:
            print(v[split])
        for item in v[split]:
            data[k].append(item)
    return data


if __name__ == "__main__":
    split_by_p = SplitIndexer(100)
    print(split_by_p.default_split_names)
    print(split_by_p.shuffled_indices())
    print(split_by_p.select_shuffled_split_indices(Split.Training))
    a = split_by_p.select_shuffled_split_indices(None)
    print(a, len(a))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 18/03/2020
           """
__all__ = ["get_host_ip"]


def get_host_ip() -> str:
    """Get host ip.

  Returns:
  str: The obtained ip. UNKNOWN if failed."""
    import socket

    try:
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(("8.8.8.8", 80))
        ip = s.getsockname()[0]
    except:
        ip = "UNKNOWN"
    finally:
        s.close()

    return ip


if __name__ == "__main__":
    HOST_IP = get_host_ip()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/06/2020
           """

__all__ = ["Meter", "AverageMeter"]


class Meter:
    """Stores current value"""

    def __init__(self):
        self.reset()

    def reset(self):
        """"""
        self.val = 0
        self.sum = 0
        self.count = 0

    def update(self, val, n: int = 1):
        """

    :param val:
    :type val:
    :param n:
    :type n:"""
        self.val = val
        self.sum += val * n
        self.count += n


class AverageMeter(Meter):
    """Computes and stores the average and current value"""

    def reset(self):
        """"""
        super().reset()
        self.avg = 0

    def update(self, val, n: int = 1):
        """

    :param val:
    :type val:
    :param n:
    :type n:"""
        super().update(val, n)
        self.avg = self.sum / self.count



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"

from pathlib import Path

import matplotlib.image as mpimg
import numpy
import scipy.misc
from matplotlib import pyplot

__all__ = ["display_depth_map_3d"]


def display_depth_map_3d(
    data_set_directory: Path = Path.home()
    / "Datasets"
    / "neodroid"
    / "depth"
    / "80.png",
):
    """

  :param data_set_directory:
  :type data_set_directory:"""
    # lena = scipy.misc.ascent()

    # downscaling has a 'smoothing' effect
    # lena = scipy.misc.imresize(lena, 0.15, interp='cubic')

    img = mpimg.imread(data_set_directory)
    img = img[..., 0]

    # def ivas(x, cam_ang):
    #  return x * math.cos(math.radians(90 - cam_ang))

    # image_length = img.shape[0]
    # camera_angle = 45.
    # ys = numpy.array([ivas(x, camera_angle) / 255 for x in range(0, image_length)])
    # ys = numpy.repeat(ys, img.shape[0], axis=0).reshape((img.shape[0], img.shape[0]))
    # img = img - ys

    img = scipy.misc.imresize(img, 0.2, interp="cubic")

    # img = img/255.

    # create the x and y coordinate arrays (here we just use pixel indices)
    d0 = img.shape[0]
    d1 = img.shape[1] / 2
    # xx, yy = numpy.mgrid[-d0:d0, -d1:d1]
    xx, yy = numpy.mgrid[0:d0, -d1:d1]

    fig = pyplot.figure()
    ax = fig.gca(projection="3d")
    ax.plot_surface(xx, yy, img, rstride=1, cstride=1, cmap=pyplot.cm.gray, linewidth=0)

    pyplot.show()


if __name__ == "__main__":
    display_depth_map_3d()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/03/2020
           """

__all__ = ["evaluate_context"]

from typing import Any, Callable, Tuple


def evaluate_context(x: Any, *args, **kwargs) -> Tuple:
    """

  :param x:
  :param args:
  :param kwargs:
  :return:"""
    if isinstance(x, Callable):
        x = x(*args, **kwargs)
    return (
        [evaluate_context(a) for a in args],
        {k: evaluate_context(v) for k, v in kwargs.items()},
        x,
        type(x),
    )


if __name__ == "__main__":

    print(evaluate_context(2, 2))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/08/2020
           """

__all__ = []


def test_import():
    import draugr

    print(draugr.__version__)


def test_import_torch_utilities():
    import draugr.torch_utilities

    print(draugr.torch_utilities.__author__)


def test_import_drawers():
    import draugr.drawers

    print(draugr.drawers.__author__)


def test_import_opencv_utilities():
    import draugr.opencv_utilities

    print(draugr.opencv_utilities.__author__)


def test_import_entry_points():
    import draugr.entry_points

    print(draugr.entry_points.__author__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

from .atari_weight_init import *
from .fan_in_weight_init import *

from .normal_weight_init import *
from .ortho_weight_init import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Any, Iterable, Sequence

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/10/2019
           """

import numpy

__all__ = ["sized_batch", "shuffled_batches", "random_batches", "batch_generator"]


def sized_batch(sized: Iterable, n: int = 32, drop_not_full: bool = True) -> Any:
    r"""

  :param sized:
  :param n:
  :param drop_not_full:
  :return:"""
    if not isinstance(sized, Sequence):
        sized = list(sized)
    l = len(sized)
    for ndx in range(0, l, n):
        if drop_not_full and ndx + n > l - 1:
            return
        yield sized[ndx : min(ndx + n, l)]


def random_batches(*args, size: int, batch_size: int) -> Sequence:
    r"""

  :param args:
  :type args:
  :param size:
  :type size:
  :param batch_size:
  :type batch_size:"""
    for _ in range(size // batch_size):
        rand_ids = numpy.random.randint(0, size, batch_size)
        yield [a[rand_ids] for a in args]


def shuffled_batches(*args, size: int, batch_size: int) -> Sequence:
    r"""

  :param args:
  :type args:
  :param size:
  :type size:
  :param batch_size:
  :type batch_size:"""
    permutation = numpy.random.permutation(size)
    r = size // batch_size
    for i in range(r):
        perm = permutation[i * batch_size : (i + 1) * batch_size]
        yield [a[perm] for a in args]


def batch_generator(iterable: Iterable, n: int = 32, drop_not_full: bool = True) -> Any:
    r"""

  :param iterable:
  :param n:
  :param drop_not_full:
  :return:"""
    b = []
    i = 0
    for a in iterable:
        b.append(a)
        i += 1
        if i >= n:
            yield b
            b.clear()
            i = 0
    if drop_not_full:
        return
    return b


if __name__ == "__main__":

    def asda():
        arg_num = 4
        size = 12
        mini_batch_size = 5
        b = numpy.random.random((arg_num, size))
        for a in shuffled_batches(*b, size=size, batch_size=mini_batch_size):
            print(list(a))

    asda()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

try:
    from .csv_writer import *
    from .log_writer import *
    from .mixins import *
    from .mock_writer import *
    from .terminal import *
    from .writer import *
    from .standard_tags import *
except ImportError as ix:
    print(
        f"Make sure requirements is installed for {pathlib.Path(__file__).parent.name}"
    )
    raise ix



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

__doc__ = r"""
"""

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

try:
    from draugr.drawers.mpl_drawers.discrete_scroll_plot import *
    from draugr.drawers.mpl_drawers.mpldrawer import *
    from draugr.drawers.mpl_drawers.series_scroll_plot import *
    from draugr.drawers.mpl_drawers.spectral import *
except ImportError as ix:
    print(
        f"Make sure requirements is installed for {pathlib.Path(__file__).parent.name}"
    )
    raise ix



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"
"""
Description: Visualisation
Author: Christian Heider Nielsen
"""
import numpy

__all__ = ["update_visualiser"]


def update_visualiser(
    visualiser,
    episode,
    loss,
    moving_loss,
    signal,
    moving_reward,
    episode_length,
    moving_length,
    rgb_array,
    windows,
    configuration,
):
    """

  :param moving_loss:
  :param signal:
  :param moving_reward:
  :param moving_length:
  :param configuration:
  :param visualiser:
  :param episode:
  :param loss:
  :param episode_length:
  :param rgb_array:
  :param windows:
  :return:"""
    if "loss" in windows:
        loss_window = windows["loss"]
        visualiser.line(
            X=numpy.array([episode]),
            Y=numpy.array([moving_loss]),
            win=loss_window,
            env=configuration.CONFIG_NAME,
            update="append",
        )
    else:
        windows["loss"] = visualiser.line(
            X=numpy.array([episode]),
            Y=numpy.array([moving_loss]),
            env=configuration.CONFIG_NAME,
            opts={"title": "Average Episode Q Value Loss"},
        )

    if "signal" in windows:
        reward_window = windows["signal"]
        visualiser.line(
            X=numpy.array([episode]),
            Y=numpy.array([moving_reward]),
            win=reward_window,
            env=configuration.CONFIG_NAME,
            update="append",
        )
    else:
        windows["signal"] = visualiser.line(
            X=numpy.array([episode]),
            Y=numpy.array([moving_reward]),
            env=configuration.CONFIG_NAME,
            opts={"title": "Average Episode Reward"},
        )

    if "episode_length" in windows:
        episode_window = windows["episode_length"]
        visualiser.line(
            X=numpy.array([episode]),
            Y=numpy.array([moving_length]),
            win=episode_window,
            env=configuration.CONFIG_NAME,
            update="append",
        )
    else:
        windows["episode_length"] = visualiser.line(
            X=numpy.array([episode]),
            Y=numpy.array([moving_length]),
            env=configuration.CONFIG_NAME,
            opts={"title": "Episode Length"},
        )

    # if 'rgb_array' in windows:
    #  rgb_array_window = windows['rgb_array']
    #  visualiser.image(rgb_array, win=rgb_array_window,
    #                   env=configuration.CONFIG_NAME)
    # else:
    #  windows['rgb_array'] = visualiser.image(rgb_array,
    #                                          env=configuration.CONFIG_NAME,
    #                                          opts={'title'  : 'Episode Ending',
    #                                                'caption': 'Episode
    # Ending'})

    return windows



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 10/12/2019
           """

from typing import Sequence

from matplotlib import animation, pyplot

__all__ = ["replay_frames"]


def replay_frames(
    frames: Sequence, interval: int = 100, is_ipython: bool = False
) -> None:
    """
  Displays a list of frames as a gif, with controls"""
    # pyplot.figure(figsize=(frames[0].shape[1] / 72.0, frames[0].shape[0] / 72.0), dpi = 72)
    patch = pyplot.imshow(frames[0])
    pyplot.axis("off")

    def animate(start_episode):
        patch.set_data(frames[start_episode])

    anim = animation.FuncAnimation(
        pyplot.gcf(), animate, frames=len(frames), interval=interval
    )
    if is_ipython:
        from IPython.display import display

        display(anim)
    else:
        pyplot.show()


if __name__ == "__main__":

    def main():
        import gym

        env = gym.make("Pendulum-v0")
        state = env.reset()

        frames = []
        done = False
        while not done:
            frames.append(env.render())

            state, reward, done, info = env.step(env.action_space.sample())
        env.close()

        replay_frames(frames)

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod
from typing import Mapping, Sequence

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["SpectrogramWriterMixin"]


class SpectrogramWriterMixin(ABC):
    """
  Writer mixin that provides an interface for 'writing' spectrogram charts"""

    @abstractmethod
    def spectrogram(
        self,
        tag: str,
        values: list,
        sample_rate: int,
        step: int,
        NFFT: int = 512,
        x_labels: Sequence = None,
        y_label: str = "Magnitude",
        x_label: str = "Sequence",
        plot_kws: Mapping = {},  # Seperate as parameters name collisions might occur
        **kwargs
    ) -> None:
        """

    :param tag:
    :type tag:
    :param data:
    :type data:
    :param step:
    :type step:
    :param dataformats:
    :type dataformats:
    :param kwargs:
    :type kwargs:"""
        raise NotImplementedError

    '''
@abstractmethod
def mfcc_spectrogram(self,
          tag: str,
          values: list,
          sample_rate:int,
          step: int,
          NFFT:int=512,
          x_labels: Sequence = None,
          y_label: str = "Magnitude",
          x_label: str = "Sequence",
          plot_kws: Mapping = {},  # Seperate as parameters name collisions might occur
          **kwargs) -> None:
"""

:param tag:
:type tag:
:param data:
:type data:
:param step:
:type step:
:param dataformats:
:type dataformats:
:param kwargs:
:type kwargs:
"""
raise NotImplementedError
'''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Iterable, Any

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11/11/2019
           """

__all__ = ["yield_and_map", "inner_map", "kw_map"]


def yield_and_map(iterable: Iterable, level: int = 0, func: callable = print) -> Any:
    """

  :param iterable:
  :type iterable:
  :param level:
  :type level:
  :param func:
  :type func:"""
    if level == 0:
        for a in iterable:
            func(a)
            yield a
    elif level == 1:
        for a in iterable:
            for b in a:
                func(b)
                yield b
    elif level == 2:
        for a in iterable:
            for b in a:
                for c in b:
                    func(c)
                    yield c


def inner_map(func: callable, iterable: Iterable, aggregate_yield: bool = True) -> Any:
    """

  :param func:
  :type func:
  :param iterable:
  :type iterable:
  :param aggregate_yield:
  :type aggregate_yield:"""
    if aggregate_yield:
        for a in iterable:
            yield [func(b) for b in a]
    else:
        for a in iterable:
            for b in a:
                yield func(b)


def kw_map(func: callable, kw: str, iterable: Iterable) -> Any:
    """

  :param func:
  :type func:
  :param kw:
  :type kw:
  :param iterable:
  :type iterable:"""
    for a in iterable:
        yield func(**{kw: a})


if __name__ == "__main__":
    a = (2, 3)
    # TODO



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02-02-2021
           """

import numpy
import torch

__all__ = [
    "numpy_to_torch_dtype_dict",
    "torch_to_numpy_dtype_dict",
    "numpy_to_torch_dtype",
    "torch_to_numpy_dtype",
]

numpy_to_torch_dtype_dict = {  # Dict of NumPy dtype -> torch dtype (when the correspondence exists)
    numpy.bool: torch.bool,
    numpy.uint8: torch.uint8,
    numpy.int8: torch.int8,
    numpy.int16: torch.int16,
    numpy.int32: torch.int32,
    numpy.int64: torch.int64,
    numpy.float16: torch.float16,
    numpy.float32: torch.float32,
    numpy.float64: torch.float64,
    numpy.complex64: torch.complex64,
    numpy.complex128: torch.complex128,
}

torch_to_numpy_dtype_dict = {
    value: key for (key, value) in numpy_to_torch_dtype_dict.items()
}  # Dict of torch dtype -> NumPy dtype


def numpy_to_torch_dtype(numpy_dtype: numpy.dtype) -> torch.dtype:
    return numpy_to_torch_dtype_dict[numpy_dtype.type]


def torch_to_numpy_dtype(torch_dtype: torch.dtype) -> numpy.dtype:
    return torch_to_numpy_dtype_dict[torch_dtype]


if __name__ == "__main__":

    def iusahdu():
        a = numpy.zeros((1, 1))
        print(a.dtype)
        b = numpy_to_torch_dtype(a.dtype)
        print(b)
        print(type(b))
        c = torch_to_numpy_dtype(b)
        print(c)
        print(type(c))

    iusahdu()
